# spectroscopy.py spectroscopy measurement class for use with a signal analyzer instead of a VNA
# based on spectroscopy.py
# TW 2018

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import numpy as np
import logging
from time import sleep,time
import sys
import threading

import qkit
if qkit.module_available("matplotlib"):
    import matplotlib.pylab as plt
if qkit.module_available("scipy"):
    from scipy.optimize import curve_fit
from qkit.storage import store as hdf
from qkit.gui.plot import plot as qviewkit
from qkit.gui.notebook.Progress_Bar import Progress_Bar
from qkit.measure.measurement_class import Measurement
import qkit.measure.write_additional_files as waf


class spectrum(object):
    """
    usage: similar to the old spectroscopy file.However class generates hdf vectors/matrices depending on the number
    of traces of your signal_analyzer.

    m = spectrum(sig_analyzer = specki)

    m.set_x_parameters(arange(-0.05,0.05,0.01),'flux coil current',coil.set_current, unit = 'mA')
    m.set_y_parameters(arange(4e9,7e9,10e6),'excitation frequency',mw_src1.set_frequency, unit = 'Hz')

    m.gen_fit_function(...)      several times

    m.measure_XX()
    functions that must be additionally implemented in your signal analyzer's driver
    get_y_unit(trace) - returns y-unit of your trace
    get_active_traces() - returns number of traces
    get_trace_name(trace) - returns name of your trace
    """

    def __init__(self, sig_analyzer, exp_name='', sample=None):

        self.sig_analyzer = sig_analyzer
        self.averaging_start_ready = "start_measurement" in self.sig_analyzer.get_function_names() \
                                     and "ready" in self.sig_analyzer.get_function_names()
        if not self.averaging_start_ready: logging.warning(
            __name__ + ': With your signal analyzer driver (' + self.sig_analyzer.get_type() + '),'
                        ' I can not see when a measurement is complete. '
                        'So I only wait for a specific time and hope the VNA has finished. '
                        'Please consider implemeting the necessary functions into your driver.')
        self.exp_name = exp_name
        self._sample = sample

        self.landscape = None
        self.span = 200e6  # [Hz]
        self.tdx = 0.002  # [s]
        self.tdy = 0.002  # [s]

        self.comment = ''
        self.dirname = None

        self.x_set_obj = None
        self.y_set_obj = None

        self.progress_bar = True
        self._plot_comment = ""

        self.set_log_function()

        self.open_qviewkit = True
        self.qviewkit_singleInstance = False

        self._measurement_object = Measurement()
        self._measurement_object.measurement_type = 'spectroscopy'
        self._measurement_object.sample = self._sample

        self._qvk_process = False


    def set_log_function(self, func=None, name=None, unit=None, log_dtype=None):
        '''
        A function (object) can be passed to the measurement loop which is excecuted before every x iteration
        but after executing the x_object setter in 2D measurements and before every line (but after setting
        the x value) in 3D measurements.
        The return value of the function of type float or similar is stored in a value vector in the h5 file.

        Call without any arguments to delete all log functions. The timestamp is automatically saved.

        func: function object in list form
        name: name of logging parameter appearing in h5 file, default: 'log_param'
        unit: unit of logging parameter, default: ''
        log_dtype: h5 data type, default: 'f' (float32)
        '''

        if name == None:
            try:
                name = ['log_param'] * len(func)
            except Exception:
                name = None
        if unit == None:
            try:
                unit = [''] * len(func)
            except Exception:
                unit = None
        if log_dtype == None:
            try:
                log_dtype = ['f'] * len(func)
            except Exception:
                log_dtype = None

        self.log_function = []
        self.log_name = []
        self.log_unit = []
        self.log_dtype = []

        if func != None:
            for i, f in enumerate(func):
                self.log_function.append(f)
                self.log_name.append(name[i])
                self.log_unit.append(unit[i])
                self.log_dtype.append(log_dtype[i])

    def set_x_parameters(self, x_vec, x_coordname, x_set_obj, x_unit=""):
        '''
        Sets parameters for sweep. In a 3D measurement, the x-parameters will be the "outer" sweep.
        For every x value all y values are swept
        Input:
        x_vec (array): conains the sweeping values
        x_coordname (string)
        x_instrument (obj): callable object to execute with x_vec-values (i.e. vna.set_power())
        x_unit (string): optional
        '''
        self.x_vec = x_vec
        self.x_coordname = x_coordname
        self.x_set_obj = x_set_obj
        self.delete_fit_function()
        self.x_unit = x_unit

    def set_y_parameters(self, y_vec, y_coordname, y_set_obj, y_unit=""):
        '''
        Sets parameters for sweep. In a 3D measurement, the x-parameters will be the "outer" sweep.
        For every x value all y values are swept
        Input:
        y_vec (array): contains the sweeping values
        y_coordname (string)
        y_instrument (obj): callable object to execute with x_vec-values (i.e. vna.set_power())
        y_unit (string): optional
        '''
        self.y_vec = y_vec
        self.y_coordname = y_coordname
        self.y_set_obj = y_set_obj
        self.delete_fit_function()
        self.y_unit = y_unit

    def gen_fit_function(self, curve_f, curve_p, units='', p0=[-1, 0.1, 7]):
        '''
        curve_f: 'parab', 'hyp', specifies the fit function to be employed
        curve_p: set of points that are the basis for the fit in the format [[x1,x2,x3,...],[y1,y2,y3,...]], frequencies in Hz
        units: set this to 'Hz' in order to avoid large values that cause the fit routine to diverge
        p0 (optional): start parameters for the fit, must be an 1D array of length 3 ([a,b,c,d]),
           where for the parabula p0[3] will be ignored
        The parabolic function takes the form  y = a*(x-b)**2 + c , where (a,b,c) = p0
        The hyperbolic function takes the form y = sqrt[ a*(x-b)**2 + c ], where (a,b,c) = p0

        adds a trace to landscape
        '''
        if not qkit.module_available("scipy"):
            raise ImportError('scipy not available.')

        if not self.landscape:
            self.landscape = []

        x_fit = curve_p[0]
        if units == 'Hz':
            y_fit = np.array(curve_p[1]) * 1e-9
        else:
            y_fit = np.array(curve_p[1])

        try:
            multiplier = 1
            if units == 'Hz':
                multiplier = 1e9

            fit_fct = None
            if curve_f == 'parab':
                fit_fct = self.f_parab
            elif curve_f == 'hyp':
                fit_fct = self.f_hyp
            elif curve_f == 'lin':
                fit_fct = self.f_lin
                p0 = p0[:2]
            else:
                print('function type not known...aborting')
                raise ValueError

            popt, pcov = curve_fit(fit_fct, x_fit, y_fit, p0=p0)
            self.landscape.append(multiplier * fit_fct(self.x_vec, *popt))

        except Exception as message:
            print('fit not successful:', message)
            popt = p0

    def _prepare_measurement_sig_analyzer(self):
        '''
        all the relevant settings from the vna are updated and called
        '''

        self.sig_analyzer.get_all()
        # ttip.get_temperature()
        self._nop = self.sig_analyzer.get_nop()
        self._sweeptime_averages = self.sig_analyzer.get_sweeptime_averages()
        self._freqpoints = self.sig_analyzer.get_freqpoints()
        self.num_traces = self.sig_analyzer.get_active_traces()
        self.traces_names = []
        self.units = []
        for i in range(self.num_traces):
            self.traces_names.append(self.sig_analyzer.get_trace_name(i+1))  # must be implemented in driver
            self.units.append(self.sig_analyzer.get_y_unit(i+1))  # must also be impelmented

        if self.averaging_start_ready:
            self.sig_analyzer.pre_measurement()

    def _prepare_measurement_file(self):
        '''
        creates the output .h5-file with distinct dataset structures for each measurement type.
        at this point all measurement parameters are known and put in the output file
        '''

        self._data_file = hdf.Data(name=self._file_name, mode='a')
        self._measurement_object.uuid = self._data_file._uuid
        self._measurement_object.hdf_relpath = self._data_file._relpath
        self._measurement_object.instruments = qkit.instruments.get_instrument_names()

        self._measurement_object.save()
        self._mo = self._data_file.add_textlist('measurement')
        self._mo.append(self._measurement_object.get_JSON())

        # write logfile and instrument settings
        self._write_settings_dataset()
        self._log = waf.open_log_file(self._data_file.get_filepath())
        self._data_freq = self._data_file.add_coordinate('frequency', unit='Hz')
        self._data_freq.add(self._freqpoints)
        self._data = []  # empty list as we have a variable number of channels

        if self._scan_1D:
            for i in range(self.num_traces):
                self._data.append(self._data_file.add_value_vector(self.traces_names[i], x=self._data_freq,
                                                                   unit=self.units[i], save_timestamp=True))
        if self._scan_2D:
            self._data_x = self._data_file.add_coordinate(self.x_coordname, unit=self.x_unit)
            self._data_x.add(self.x_vec)
            for i in range(self.num_traces):
                self._data.append(self._data_file.add_value_matrix(self.traces_names[i], x=self._data_x,
                                                                   y=self._data_freq, unit=self.units[i],
                                                                   save_timestamp=True))
            if self.log_function != None:  # use logging
                self._log_value = []
                for i in range(len(self.log_function)):
                    self._log_value.append(
                        self._data_file.add_value_vector(self.log_name[i], x=self._data_x, unit=self.log_unit[i],
                                                         dtype=self.log_dtype[i]))

        if self._scan_3D:
            self._data_x = self._data_file.add_coordinate(self.x_coordname, unit=self.x_unit)
            self._data_x.add(self.x_vec)
            self._data_y = self._data_file.add_coordinate(self.y_coordname, unit=self.y_unit)
            self._data_y.add(self.y_vec)
            for i in range(self.num_traces):
                self._data.append(self._data_file.add_value_box(self.traces_names[i], x=self._data_x, y = self._data_y,
                                                                   z=self._data_freq, unit=self.units[i],
                                                                   save_timestamp=True))
            if self.log_function != None:  # use logging
                self._log_value = []
                for i in range(len(self.log_function)):
                    self._log_value.append(
                        self._data_file.add_value_vector(self.log_name[i], x=self._data_x, unit=self.log_unit[i],
                                                         dtype=self.log_dtype[i]))

        if self.comment:
            self._data_file.add_comment(self.comment)

        if self.qviewkit_singleInstance and self.open_qviewkit and self._qvk_process:
            self._qvk_process.terminate()  # terminate an old qviewkit instance

    def _write_settings_dataset(self):
        self._settings = self._data_file.add_textlist('settings')
        settings = waf.get_instrument_settings(self._data_file.get_filepath())
        self._settings.append(settings)

    def measure_1D(self, rescan=True, web_visible=True):
        '''
        measure method to record a single (averaged) VNA trace, S11 or S21 according to the setting on the VNA
        rescan: If True (default), the averages on the VNA are cleared and a new measurement is started.
                If False, it will directly take the data from the VNA without waiting.
        '''
        self._scan_1D = True
        self._scan_2D = False
        self._scan_3D = False
        self._scan_time = False

        self._measurement_object.measurement_func = 'measure_1D'
        self._measurement_object.x_axis = 'frequency'
        self._measurement_object.y_axis = ''
        self._measurement_object.z_axis = ''
        self._measurement_object.web_visible = web_visible

        if not self.dirname:
            self.dirname = 'signal_analyzer_tracedata'
        self._file_name = self.dirname.replace(' ', '').replace(',', '_')
        if self.exp_name:
            self._file_name += '_' + self.exp_name
        self._prepare_measurement_sig_analyzer()
        self._prepare_measurement_file()

        """opens qviewkit to plot measurement, amp and pha are opened by default"""
        if self.open_qviewkit:
            self._qvk_process = qviewkit.plot(self._data_file.get_filepath(), datasets=self.traces_names)
        print('recording trace...')
        sys.stdout.flush()

        qkit.flow.start()
        if rescan:
            if self.averaging_start_ready:
                self.sig_analyzer.start_measurement()
                ti = time()
                if self.progress_bar: self._p = Progress_Bar(self.sig_analyzer.get_averages(), self.dirname,
                                                             self.sig_analyzer.get_sweeptime())
                qkit.flow.sleep(.2)
                while not self.sig_analyzer.ready():
                    if time() - ti > self.sig_analyzer.get_sweeptime(query=False):
                        if self.progress_bar: self._p.iterate()
                        ti = time()
                    qkit.flow.sleep(.2)
                if self.progress_bar:
                    while self._p.progr < self._p.max_it:
                        self._p.iterate()
            else:  # not tested!!!!!!
                self.sig_analyzer.avg_clear()
                if self.sig_analyzer.get_averages() == 1 or self.sig_analyzer.get_Average() == False:  # no averaging
                    if self.progress_bar: self._p = Progress_Bar(1, self.dirname, self.sig_analyzer.get_sweeptime())
                    qkit.flow.sleep(self.sig_analyzer.get_sweeptime())  # wait single sweep
                    if self.progress_bar: self._p.iterate()
                else:  # with averaging
                    if self.progress_bar: self._p = Progress_Bar(self.sig_analyzer.get_averages(), self.dirname,
                                                                 self.sig_analyzer.get_sweeptime())
                    if "avg_status" in self.sig_analyzer.get_function_names():
                        for a in range(self.sig_analyzer.get_averages()):
                            while self.sig_analyzer.avg_status() <= a:
                                qkit.flow.sleep(.2)  # maybe one would like to adjust this at a later point
                            if self.progress_bar: self._p.iterate()
                    else:  # old style
                        for a in range(self.sig_analyzer.get_averages()):
                            qkit.flow.sleep(self.sig_analyzer.get_sweeptime())  # wait single sweep time
                            if self.progress_bar: self._p.iterate()

        for i in range(self.num_traces):
            self._data[i].append(self.sig_analyzer.get_tracedata(i+1))

        qkit.flow.end()
        self._end_measurement()

    def measure_2D(self, web_visible=True):
        '''
        measure method to record a (averaged) VNA trace, S11 or S21 according to the setting on the VNA
        for all parameters x_vec in x_obj
        '''

        if not self.x_set_obj:
            logging.error('axes parameters not properly set...aborting')
            return
        self._scan_1D = False
        self._scan_2D = True
        self._scan_3D = False
        self._scan_time = False

        self._measurement_object.measurement_func = 'measure_2D'
        self._measurement_object.x_axis = self.x_coordname
        self._measurement_object.y_axis = 'frequency'
        self._measurement_object.z_axis = ''
        self._measurement_object.web_visible = web_visible

        if not self.dirname:
            self.dirname = self.x_coordname
        self._file_name = '2D_' + self.dirname.replace(' ', '').replace(',', '_')
        if self.exp_name:
            self._file_name += '_' + self.exp_name

        if self.progress_bar: self._p = Progress_Bar(len(self.x_vec), '2D signal analyzer sweep ' + self.dirname,
                                                     self.sig_analyzer.get_sweeptime_averages())

        self._prepare_measurement_sig_analyzer()
        self._prepare_measurement_file()

        """opens qviewkit to plot measurement, amp and pha are opened by default"""

        if self.open_qviewkit: self._qvk_process = qviewkit.plot(self._data_file.get_filepath(),
                                                                 datasets=self.traces_names)
        self._measure()

    def measure_3D(self, web_visible=True):
        '''
        measure full window of vna while sweeping x_set_obj and y_set_obj with parameters x_vec/y_vec.
        sweep over y_set_obj is the inner loop, for every value x_vec[i] all values y_vec are measured.

        optional: measure method to perform the measurement according to landscape, if set
        self.span is the range (in units of the vertical plot axis) data is taken around the specified funtion(s)
        note: make sure to have properly set x,y vectors before generating traces
        '''
        if not self.x_set_obj or not self.y_set_obj:
            logging.error('axes parameters not properly set...aborting')
            return
        self._scan_1D = False
        self._scan_2D = False
        self._scan_3D = True
        self._scan_time = False

        self._measurement_object.measurement_func = 'measure_3D'
        self._measurement_object.x_axis = self.x_coordname
        self._measurement_object.y_axis = self.y_coordname
        self._measurement_object.z_axis = 'frequency'
        self._measurement_object.web_visible = web_visible

        if not self.dirname:
            self.dirname = self.x_coordname + ', ' + self.y_coordname
        self._file_name = '3D_' + self.dirname.replace(' ', '').replace(',', '_')
        if self.exp_name:
            self._file_name += '_' + self.exp_name

        if self.progress_bar: self._p = Progress_Bar(len(self.x_vec) * len(self.y_vec), '3D signal analyzer sweep ' +
                                                     self.dirname, self.sig_analyzer.get_sweeptime_averages())
        self._prepare_measurement_sig_analyzer()
        self._prepare_measurement_file()
        """opens qviewkit to plot measurement, amp and pha are opened by default"""
        """only middle point in freq array is plotted vs x and y"""
        if self.open_qviewkit: self._qvk_process = qviewkit.plot(self._data_file.get_filepath(),
                                                                 datasets = self.traces_names)

        ### ???
        if self.landscape:
            self.center_freqs = np.array(self.landscape).T
        else:
            self.center_freqs = []  # load default sequence
            for i in range(len(self.x_vec)):
                self.center_freqs.append([0])

        self._measure()


    def _measure(self):
        '''
        measures and plots the data depending on the measurement type.
        the measurement loops feature the setting of the objects and saving the data in the .h5 file.
        '''
        qkit.flow.start()
        try:
            """
            loop: x_obj with parameters from x_vec
            """
            for ix, x in enumerate(self.x_vec):
                self.x_set_obj(x)
                sleep(self.tdx)

                if self.log_function != None:
                    for i, f in enumerate(self.log_function):
                        self._log_value[i].append(float(f()))

                #### not tested
                if self._scan_3D:
                    for y in self.y_vec:
                        """
                        loop: y_obj with parameters from y_vec (only 3D measurement)
                        """
                        if (np.min(np.abs(self.center_freqs[ix] - y * np.ones(len(self.center_freqs[
                                                                                      ix])))) > self.span / 2.) and self.landscape:  # if point is not of interest (not close to one of the functions)
                            data = []
                            for i in range(self.num_traces):
                                data.append(np.zeros(int(self._nop)))
                        else:
                            self.y_set_obj(y)
                            sleep(self.tdy)
                            if self.averaging_start_ready:
                                self.sig_analyzer.start_measurement()
                                qkit.flow.sleep(.2)  # just to make sure, the ready command does not *still* show ready
                                while not self.sig_analyzer.ready():
                                    qkit.flow.sleep(.2)
                            else:
                                self.sig_analyzer.avg_clear()
                                qkit.flow.sleep(self._sweeptime_averages)

                            # if "avg_status" in self.sig_analyzer.get_function_names():
                            #       while self.sig_analyzer.avg_status() < self.sig_analyzer.get_averages():
                            #            qkit.flow.sleep(.2) #maybe one would like to adjust this at a later point

                            """ measurement """
                            for i in range(self.num_traces):
                                self._data[i].append(self.sig_analyzer.get_tracedata(i+1))

                        if self.progress_bar:
                            self._p.iterate()
                        qkit.flow.sleep()
                    """
                    filling of value-box is done here.
                    after every y-loop the data is stored the next 2d structure
                    """
                    for i in range(self.num_traces):
                        self._data[i].next_matrix()

                if self._scan_2D:
                    if self.averaging_start_ready:
                        self.sig_analyzer.start_measurement()
                        qkit.flow.sleep(.2)  # just to make sure, the ready command does not *still* show ready
                        while not self.sig_analyzer.ready():
                            qkit.flow.sleep(.2)
                    else:
                        self.sig_analyzer.avg_clear()
                        qkit.flow.sleep(self._sweeptime_averages)
                    """ measurement """
                    for i in range(self.num_traces):
                        self._data[i].append(self.sig_analyzer.get_tracedata(i+1))

                    if self.progress_bar:
                        self._p.iterate()
                    qkit.flow.sleep()
        finally:
            self._end_measurement()
            qkit.flow.end()

    def _end_measurement(self):
        '''
        the data file is closed and filepath is printed
        '''
        print(self._data_file.get_filepath())
        # qviewkit.save_plots(self._data_file.get_filepath(),comment=self._plot_comment) #old version where we have to wait for the plots
        t = threading.Thread(target=qviewkit.save_plots, args=[self._data_file.get_filepath(), self._plot_comment])
        t.start()
        self._data_file.close_file()
        waf.close_log_file(self._log)
        self.dirname = None
        if self.averaging_start_ready: self.sig_analyzer.post_measurement()


    def delete_fit_function(self, n=None):
        '''
        delete single fit function n (with 0 being the first one generated) or the complete landscape for n not specified
        '''

        if n:
            self.landscape = np.delete(self.landscape, n, axis=0)
        else:
            self.landscape = None

    def plot_fit_function(self, num_points=100):
        '''
        try:
            x_coords = np.linspace(self.x_vec[0], self.x_vec[-1], num_points)
        except Exception as message:
            print 'no x axis information specified', message
            return
        '''
        if not qkit.module_available("matplotlib"):
            raise ImportError("matplotlib not found.")
        if self.landscape:
            for trace in self.landscape:
                try:
                    # plt.clear()
                    plt.plot(self.x_vec, trace)
                    plt.fill_between(self.x_vec, trace + float(self.span) / 2, trace - float(self.span) / 2, alpha=0.5)
                except Exception:
                    print('invalid trace...skip')
            plt.axhspan(self.y_vec[0], self.y_vec[-1], facecolor='0.5', alpha=0.5)
            plt.show()
        else:
            print('No trace generated.')

    def set_span(self, span):
        self.span = span

    def get_span(self):
        return self.span

    def set_tdx(self, tdx):
        self.tdx = tdx

    def set_tdy(self, tdy):
        self.tdy = tdy

    def get_tdx(self):
        return self.tdx

    def get_tdy(self):
        return self.tdy

    def f_parab(self, x, a, b, c):
        return a * (x - b) ** 2 + c

    def f_hyp(self, x, a, b, c):
        "hyperbolic function with the form y = sqrt[ a*(x-b)**2 + c ]"
        return np.sqrt(a * (x - b) ** 2 + c)

    def f_lin(self, x, a, b):
        return a * x + b

    def set_plot_comment(self, comment):
        '''
        Small comment to add at the end of plot pics for more information i.e. good for wiki entries.
        '''
        self._plot_comment = comment



# -*- coding: utf-8 -*-

# qplexkit.py driver for the homemade qplexkit used for DC multiplexing
# Micha Wildermuth, micha.wildermuth@kit.edu 2023
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit
from qkit.services.qplexkit import qplexkit
from qkit.core.instrument_base import Instrument
from qkit.config.services import cfg
import logging
import zmq

""" copy docstrings from qplexkit """
import sys
from typing import Callable, TypeVar
if sys.version_info <= (3, 9):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


def use_docstring(wrapper: Callable[P, T]):
    def decorator(func: Callable) -> Callable[P, T]:
        func.__doc__ = wrapper.__doc__
        return func
    return decorator


class qplexkit_client(Instrument):
    """
    This is the driver for the homemade qplexkit. It is the interface to the Raspberry Pi that controls a current
    source to switch relays at cryogenic temperatures.

    Usage:
    ------
    Initialize with
    <name> = qkit.instruments.create('<name>', 'qplexkit', address=<address>, port=<port>)
    """

    def __init__(self, name, address, port=None):
        """
        Initializes zmq communication with the qplexkit server running on a Raspberry Pi to control the relays of the
        qplexkit

        Parameters
        ----------
        name: string
            Name of the instrument (driver).
        address: string
            IP-address of Raspberry Pi
        port: int, string
            Port of zeroMQ server

        Returns
        -------
        None

        Examples
        --------
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]

        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py
        Initialized the file info database (qkit.fid) in 0.000 seconds.

        >>> qpk = qkit.instruments.create('qpk', 'qplexkit', address='00.00.000.00')
        """
        self.__name__ = __name__
        # create instrument
        logging.info(f'{__name__}: Initializing instrument qplexkit')
        Instrument.__init__(self, name, tags=['physical'])
        self.cfg = cfg['qplexkit']

        ''' set up zeroMQ client '''
        self._address = address
        if port is None:
            self._port = self.cfg['server_port']
        else:
            self._port = port
        self._context = zmq.Context()
        self.socket = self._context.socket(zmq.REQ)
        self.connect()

        ''' qkit-instrument parameters & functions '''
        self.add_parameter('switch_time',
                           type=float,
                           flags=Instrument.FLAG_GETSET,
                           minval=0,
                           units='s')
        self.add_parameter('experiment',
                           type=int,
                           flags=Instrument.FLAG_GETSET,
                           minval=0,
                           maxval=11)
        self.add_parameter('relays',
                           type=list,
                           flags=Instrument.FLAG_GET)
        # self.add_parameter('current_divider',
        #                    type=bool,
        #                    flags=Instrument.FLAG_GETSET)
        # self.add_parameter('amplifier',
        #                    type=bool,
        #                    flags=Instrument.FLAG_GETSET)
        self.add_parameter('current_source_status',
                           type=bool,
                           flags=Instrument.FLAG_GETSET)
        self.add_function('set_relay')
        self.add_function('get_relay')
        self.add_function('get_ccr')
        self.add_function('read_ccr')
        self.add_function('reset')
        self.add_function('get_attr')

    def connect(self, **kwargs):
        try:
            url = f'''tcp://{kwargs.get('address', self._address)}:{kwargs.get('port', self._port)}'''
            logging.info(f'{__name__}: connecting to {url}')
            self.socket.setsockopt(zmq.RCVTIMEO, 100000)  # wait no longer than 100 second to fail.
            self.socket.setsockopt(zmq.LINGER, 0)
            self.socket.connect(f'{url}')
            if self._query(("ping", tuple([]), dict({}))) == 'pong':
                logging.info(f'{__name__}: connection to {url} successfully established')
                return True
            else:
                logging.error(f'{__name__}: connection to {url} established, but ping-pong failed. Ensure that the server is running.')
                return False
        except zmq.error.Again as e:
            url = f'''tcp://{kwargs.get('address', self._address)}:{kwargs.get('port', self._port)}'''
            logging.error(f'{__name__}: connecting to {url} failed')
            logging.error(f'{__name__}: Server not available or Authenticator failed! {e}')
            return False

    def disconnect(self, **kwargs):
        url = f'''tcp://{kwargs.get('address', self._address)}:{kwargs.get('port', self._port)}'''
        logging.info(f'{__name__}: disconnecting {url}')
        self.socket.disconnect(f'''tcp://{kwargs.get('address', self._address)}:{kwargs.get('port', self._port)}''')

    def _query(self, msg):
        """
        Sends a message <msg> and returns the read answer <ans>.

        parameters
        ----------
        msg: str
            Message that is sent to the zeroMQ server running on a Raspberry Pi.
            The message is dumped via json and encoded as binary-string to be zeroMQ compatible.

        returns
        -------
        ans: str
            Answer that is returned to the queried message <msg>.
        """
        self.socket.send_json(msg)
        ans = self.socket.recv_json()
        if type(ans) is list:
            if type(ans[0]) is str:
                if 'error' in ans[0].lower():
                    raise Exception(ans)
        return ans

    @use_docstring(qplexkit.qplexkit.set_switch_time)
    def do_set_switch_time(self, val):
        logging.info(f'''{__name__}: set switch time to {val}s''')
        msg = ("set_switch_time", tuple([val]), dict({}))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.get_switch_time)
    def do_get_switch_time(self):
        msg = ("get_switch_time", tuple([]), dict({}))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.set_experiment)
    def do_set_experiment(self, exp, protect=False, **kwargs):
        logging.info(f'''{__name__}: set experiment to {exp}''')
        msg = ("set_experiment", tuple([exp, protect]), dict(kwargs))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.get_experiment)
    def do_get_experiment(self, **kwargs):
        msg = ("get_experiment", tuple([]), dict(kwargs))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.set_current_divider)
    def do_set_current_divider(self, status, **kwargs):
        logging.info(f'''{__name__}: set current divider to {status}''')
        msg = ("set_current_divider", tuple([int(status)]), dict(kwargs))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.get_current_divider)
    def do_get_current_divider(self, **kwargs):
        msg = ("get_current_divider", tuple([]), dict(kwargs))
        return bool(self._query(msg))

    @use_docstring(qplexkit.qplexkit.set_amplifier)
    def do_set_amplifier(self, status, **kwargs):
        logging.info(f'''{__name__}: set amplifier to {status}''')
        msg = ("set_amplifier", tuple([int(status)]), dict(kwargs))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.get_amplifier)
    def do_get_amplifier(self, **kwargs):
        msg = ("get_amplifier", tuple([]), dict(kwargs))
        return bool(self._query(msg))

    @use_docstring(qplexkit.qplexkit.set_current_source_status)
    def do_set_current_source_status(self, status, **kwargs):
        logging.info(f'''{__name__}: set current source status to {status}''')
        msg = ("set_current_source_status", tuple([int(status)]), dict(kwargs))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.get_current_source_status)
    def do_get_current_source_status(self, **kwargs):
        msg = ("get_current_source_status", tuple([]), dict(kwargs))
        return bool(self._query(msg))

    @use_docstring(qplexkit.qplexkit.set_relay)
    def set_relay(self, rel, status, **kwargs):
        logging.info(f'''{__name__}: set relay {rel} to {status}''')
        msg = ("set_relay", tuple([rel, int(status)]), dict(kwargs))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.get_relay)
    def get_relay(self, rel, **kwargs):
        msg = ("get_relay", tuple([rel]), dict(kwargs))
        return bool(self._query(msg))

    @use_docstring(qplexkit.qplexkit.get_relays)
    def do_get_relays(self, **kwargs):
        msg = ("get_relays", tuple([]), dict(kwargs))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.get_ccr)
    def get_ccr(self, rel, **kwargs):
        msg = ("get_ccr", tuple([rel]), dict(kwargs))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.read_ccr)
    def read_ccr(self, n=-1, timestamp=False, **kwargs):
        msg = ("read_ccr", tuple([n, timestamp]), dict(kwargs))
        return self._query(msg)

    @use_docstring(qplexkit.qplexkit.reset)
    def reset(self):
        logging.info(f'''{__name__}: reset qplexkit by setting all relays to 0 which corresponds to experiment 0''')
        msg = ("reset", tuple([]), dict({}))
        return self._query(msg)

    def get_attr(self, attr):
        """
        Gets a class attribute <attr> of the qplexkit instance running on the Raspberry Pi.

        Parameters
        ----------
        attr: string
            Name of class attribute.

        Returns
        -------
        val:
            Value of the class attribute.
        """
        msg = ("get_attr", tuple([attr]), dict({}))
        return self._query(msg)




# Taken from the manual belonging to the joy it step motor
# and applied to work remotely with zerorpc
# rotation is threaded to give feedback during the rotation

from time import sleep
import RPi.GPIO as gpio
import numpy as np
import zerorpc
import threading

class StepMotor(object):
    """
    This class controls the step motor via GPIO Pins and
    opens a zerorpc server for connection with other pcs
    Make sure this file is executable and you start it on reboot with
    cronjob for example
    """
    def __init__(self, initial_position=0):
        gpio.setmode(gpio.BCM)

        # Pin-Assignment
        self.A = 18
        self.B = 23
        self.C = 24
        self.D = 25
        self.wait = 0.001  # this is a good value for half step_op
        self.sequence = []
        self.set_operation_mode(1)
        self.rotations = initial_position
        self.steps = 0
        self.max_steps = 0
        self.stop = False
        # defining the PINs
        gpio.setup(self.A, gpio.OUT)
        gpio.setup(self.B, gpio.OUT)
        gpio.setup(self.C, gpio.OUT)
        gpio.setup(self.D, gpio.OUT)
        gpio.output(self.A, 0)
        gpio.output(self.B, 0)
        gpio.output(self.C, 0)
        gpio.output(self.D, 0)

    # defining the 8 steps
    def step1(self):
        gpio.output(self.D, 1)
        sleep(self.wait)
        gpio.output(self.D, 0)

    def step2(self):
        gpio.output(self.D, 1)
        gpio.output(self.C, 1)
        sleep(self.wait)
        gpio.output(self.D, 0)
        gpio.output(self.C, 0)

    def step3(self):
        gpio.output(self.C, 1)
        sleep(self.wait)
        gpio.output(self.C, 0)

    def step4(self):
        gpio.output(self.C, 1)
        gpio.output(self.B, 1)
        sleep(self.wait)
        gpio.output(self.C, 0)
        gpio.output(self.B, 0)

    def step5(self):
        gpio.output(self.B, 1)
        sleep(self.wait)
        gpio.output(self.B, 0)

    def step6(self):
        gpio.output(self.B, 1)
        gpio.output(self.A, 1)
        sleep(self.wait)
        gpio.output(self.B, 0)
        gpio.output(self.A, 0)

    def step7(self):
        gpio.output(self.A, 1)
        sleep(self.wait)
        gpio.output(self.A, 0)

    def step8(self):
        gpio.output(self.A, 1)
        gpio.output(self.D, 1)
        sleep(self.wait)
        gpio.output(self.A, 0)
        gpio.output(self.D, 0)
            
    def turn(self, angle):
        self.steps = 0
        self.max_steps = int(float(angle)*512/360*27)  # 512 steps for one rotation at stepper motor times 27 due to gearbox
        rotations = self.rotations + float(angle)/360
        if rotations <= 0 or rotations >= 17.5:
            return
        else:
            self.rotations = rotations
            self.t = threading.Thread(target=self._rotate)
            self.t.start()
                
    def get_max_steps(self):
        return self.max_steps
        
    def get_steps(self):
        return self.steps
                    
    def _rotate(self):
        print 'rotating'
        if self.max_steps > 0:
            for i in np.arange(self.max_steps):
                if self.stop:
                    break
                for s in self.sequence:
                    s()
                self.steps = i
        else:
            for i in np.arange(-1*self.max_steps):
                if self.stop:
                    break
                for s in self.sequence[::-1]:
                    s()
                self.steps = i
        self.stop = False
        self.max_steps = 0

    def set_operation_mode(self, mode):
        """ 0 for full step mode (rougher but faster)
            1 for half step mode
        """
        # if mode == 0:
            # self. sequence = [self.step2, self.step4, self.step6, self.step8]
        if mode == 1:
            self.sequence = [self.step1, self.step2, self.step3, self.step4,
                             self.step5, self.step6, self.step7, self.step8]
        else:
            print 'wrong input'
                
    def stop_rotation(self):
        # ToDo: implement correct tracking in case code is used directly on raspi
        self.stop = True

    def get_operation_mode(self):
        if len(self.sequence) == 8:
            return 'half-step'
        elif len(self.sequence) == 4:
            return 'full-step'
        else:
            return 'Error'
            
    def set_waiting_time(self, waiting_time):
        """controls the speed"""
        self.wait = waiting_time

    def get_waiting_time(self):
        return self.wait

    def get_connection(self):
        return True


if __name__ == '__main__':
    s = zerorpc.Server(StepMotor())
    s.bind("tcp://0.0.0.0:4242")
    s.run()



# -*- coding: utf-8 -*-
# Micha Wildermuth, micha.wildermuth@kit.edu 2020

from qkit.core.instrument_base import Instrument
import time


class ohmmeter_dummy(Instrument)::
    '''
    This is a driver for a dummy ohmmeter as used for sputter deposition monitoring qkit.services.qdepokit.sdi.

    Usage:
        Initialize with
        <name> = qkit.instruments.create('<name>', 'mfc_dummy')
    '''
    def __init__(self, name):
        self.__name__ = __name__
        Instrument.__init__(self, name, tags=['virtual'])
        self.t0 = None
        self.add_parameter('measure_4W', type=bool,
            flags=Instrument.FLAG_GETSET)

    def do_set_measure_4W(self, status = False):
        pass

    def do_get_measure_4W(self):
        return None

    def get_resistance(self):
        if self.t0 is None:
            self.t0 = time.time()
        return 1.e9/(time.time() - self.t0 + 1e-20)


# spectroscopy.py spectroscopy measurement class for use with a vector network analyzer
# JB/MP/AS@KIT 04/2015, 08/2015, 01/2016

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import logging
from time import time

import numpy as np

import qkit

if qkit.module_available("matplotlib"):
    import matplotlib.pylab as plt
if qkit.module_available("scipy"):
    from scipy.optimize import curve_fit
    from scipy.interpolate import interp1d, UnivariateSpline
from qkit.analysis.resonator import Resonator as resonator
from qkit.gui.notebook.Progress_Bar import Progress_Bar
from qkit.measure.measurement_base import MeasureBase


##################################################################

class spectrum(MeasureBase):
    """
    Class for spectroscopy measurements with a VNA
    usage:
    m = spectrum(vna=vna1, sample=yoursample)

    m.set_x_parameters(arange(-0.05,0.05,0.01),'flux_coil_current',coil.set_current, unit = 'mA')  outer scan in 3D
    m.set_y_parameters(arange(4e9,7e9,10e6),'excitation_frequency',mw_src1.set_frequency, unit = 'Hz')

    m.landscape.generate_fit_function_xy(...) for 3D scan, can be called several times and appends the current landscape
    m.landscape.generate_fit_function_xz(...) for 2D or 3D scan, adjusts the vna freqs with respect to x

    m.measure_XX()
    """

    def __init__(self, vna, sample=None):
        super(spectrum, self).__init__(sample)
        
        self.vna = vna
        self.averaging_start_ready = "start_measurement" in self.vna.get_function_names() and "ready" in self.vna.get_function_names()
        if not self.averaging_start_ready: logging.warning(
                __name__ + ': With your VNA instrument driver (' + self.vna.get_type() + '), I can not see when a measurement is complete. So I only wait for a specific time and hope the VNA has finished. Please consider implemeting the necessary functions into your driver.')
        
        self.landscape = Landscape(vna=vna, spec=self)
        self._fit_resonator = False
        self._measurement_object.measurement_type = 'SpectroscopyMeasurement'
        self._views = []
        self._scan_time = False
        self._segments = [] # bool([]) == False
    
    def set_x_parameters(self, vec, coordname, set_obj, unit, dt=None):
        """
        Sets x-parameters for 2D and 3D scan.
        In a 3D measurement, the x-parameters will be the "outer" sweep meaning for every x value all y values are swept and for each (x,y) value the bias is swept according to the set sweep parameters.

        Parameters
        ----------
        vec: array_likes
            An N-dimensional array that contains the sweep values.
        coordname: string
            The coordinate name to be created as data series in the .h5 file.
        set_obj: obj
            An callable object to execute with vec-values.
        unit: string
            The unit name to be used in data series in the .h5 file.
        dt: float, optional
            The sleep time between x-iterations.

        Returns
        -------
        None
        """
        super(spectrum, self).set_x_parameters(vec, coordname, set_obj, unit, dt)
        if self.landscape.xylandscapes:
            self.landscape.delete_landscape_function_xy()
            logging.warning('xy landscape has been deleted')
        if self.landscape.xzlandscape_func:
            self.landscape.delete_landscape_function_xz()
            logging.warning('xz landscape has been deleted')
    
    def set_y_parameters(self, vec, coordname, set_obj, unit, dt=None):
        """
        Sets y-parameters for 2D and 3D scan.
        In a 3D measurement, the y-parameters will be the "inner" sweep meaning for every x value all y values are swept

        Parameters
        ----------
        vec: array_likes
            An N-dimensional array that contains the sweep values.
        coordname: string
            The coordinate name to be created as data series in the .h5 file.
        set_obj: obj
            An callable object to execute with vec-values.
        unit: string
            The unit name to be used in data series in the .h5 file.
        dt: float, optional
            The sleep time between y-iterations.

        Returns
        -------
        None
        """
        super(spectrum, self).set_y_parameters(vec, coordname, set_obj, unit, dt)
        if self.landscape.xylandscapes:
            self.landscape.delete_landscape_function_xy()
            logging.warning('xy landscape has been deleted')
        if self.landscape.xzlandscape_func:
            self.landscape.delete_landscape_function_xz()
            logging.warning('xz landscape has been deleted. You can switch this off if you whish.')
    
    def _prepare_measurement_devices(self):
        """
        all the relevant settings from the vna are updated and called
        """
        self.vna.get_all()
        self._nop = self.vna.get_nop()
        self._sweeptime_averages = self.vna.get_sweeptime_averages()
        try:
            self._segments = self.vna.get_segments()
        except AttributeError:
            self._segments = []
        if self._dim == 1 or not self.landscape.xzlandscape_func:  # normal scan
            self._freqpoints = self.vna.get_freqpoints()
        else:
            self._freqpoints = self.landscape.get_freqpoints_xz()
        if self.averaging_start_ready:
            self.vna.pre_measurement()

    # def _prepare_measurement_file_(self):
    #
    #     if not self._scan_time:
    #         self._data_freq = self._data_file.add_coordinate('frequency', unit='Hz')
    #         self._data_freq.add(self._freqpoints)
    #
    #     if self._scan_2D:
    #         if self._nop < 10:
    #             """creates view: plot middle point vs x-parameter, for qubit measurements"""
    #             self._views = [self._data_file.add_view("amplitude_midpoint", x=self._data_x, y=self._data_amp,
    #                                                     view_params=dict(transpose=True, default_trace=self._nop // 2, linecolors=[(200, 200, 100)])),
    #                            self._data_file.add_view("phase_midpoint", x=self._data_x, y=self._data_pha,
    #                                                     view_params=dict(transpose=True, default_trace=self._nop // 2, linecolors=[(200, 200, 100)]))]
    #
    #     if self._scan_time:
    #         self._data_x = self._data_file.add_coordinate('trace_number', unit='')
    #         self._data_x.add(np.arange(0, self.number_of_timetraces, 1))
    #
    #         self._data_amp = self._data_file.add_value_matrix('amplitude', x=self._data_x, y=self._data_time, unit='lin. mag.', save_timestamp=False)
    #         self._data_pha = self._data_file.add_value_matrix('phase', x=self._data_x, y=self._data_time, unit='rad.', save_timestamp=False)
    
    def measure_1D(self, rescan=True):
        """
        measure method to record a single (averaged) VNA trace, S11 or S21 according to the setting on the VNA
        rescan: If True (default), the averages on the VNA are cleared and a new measurement is started.
                If False, it will directly take the data from the VNA without waiting.
        """
        self._dim =1
        self._measurement_object.measurement_func = 'measure_1D'
        
        self._prepare_measurement_devices()
        if self._segments:
            f =  [self.Coordinate('frequency_%i' % i, unit='Hz', values=self._freqpoints[0 if i == 0 else self._segments[i - 1]:self._segments[i]]) for i in range(len(self._segments))]
            self._prepare_measurement_file(
                    [self.Data(P[0] % i, [f[i]], P[1], save_timestamp=P[2] and not i)
                     for i in range(len(self._segments)) for P in [["amplitude_%i","arb. unit",True],["phase_%i","rad",False],["real_%i","",False],["imag_%i","",False]]])
        else:
            f = self.Coordinate('frequency', unit='Hz', values=self._freqpoints)
    
            self._prepare_measurement_file([self.Data("real", [f], ""), self.Data("imag", [f], ""), self.Data("amplitude", [f], "arb. unit", save_timestamp=True),
                self.Data("phase", [f], "rad")])
        
        self._open_qviewkit(datasets=[] if len(self._segments)>4 else None)
        
        qkit.flow.start()
        if rescan:
            self._pb = Progress_Bar(self.vna.get_averages(), self.measurement_name, self.vna.get_sweeptime(), dummy=not self.progress_bar)
            if self.averaging_start_ready:
                self.vna.start_measurement()
                ti = time()
                qkit.flow.sleep(.2)
                while not self.vna.ready():
                    if time() - ti > self.vna.get_sweeptime(query=False):
                        self._pb.iterate()
                        ti = time()
                    qkit.flow.sleep(.2)
                while self._pb.progr < self._pb.max_it:
                    self._pb.iterate()
            else:
                self.vna.avg_clear()
                if self.vna.get_averages() == 1 or self.vna.get_Average() == False:  # no averaging
                    qkit.flow.sleep(self.vna.get_sweeptime())  # wait single sweep
                    self._pb.iterate()
                else:  # with averaging
                    if "avg_status" in self.vna.get_function_names():
                        for a in range(self.vna.get_averages()):
                            while self.vna.avg_status() <= a:
                                qkit.flow.sleep(.2)  # maybe one would like to adjust this at a later point
                            self._pb.iterate()
                    else:  # old style
                        for a in range(self.vna.get_averages()):
                            qkit.flow.sleep(self.vna.get_sweeptime())  # wait single sweep time
                            self._pb.iterate()
        
        data_amp, data_pha = self.vna.get_tracedata()
        data_real, data_imag = self.vna.get_tracedata('RealImag')
        
        self._append(data_amp,data_pha,data_real,data_imag)
        if self._fit_resonator:
            self._do_fit_resonator()
        self._end_measurement()
    
    def measure_2D(self):
        """
        measure method to record a (averaged) VNA trace, S11 or S21 according to the setting on the VNA
        for all parameters x_vec in x_obj
        """
        
        if self.landscape.xzlandscape_func:  # The vna limits need to be adjusted, happens in the frequency wrapper
            self._x_parameter.set_function = self.landscape.vna_frequency_wrapper(self._x_parameter.set_function)
        
        self._dim = 2
        self._measurement_object.measurement_func = 'measure_2D'
        
        self._prepare_measurement_devices()
        if self._segments:
            f =  [self.Coordinate('frequency_%i' % i, unit='Hz', values=self._freqpoints[0 if i == 0 else self._segments[i - 1]:self._segments[i]]) for i in range(len(self._segments))]
            self._prepare_measurement_file(
                    [self.Data(P[0] % i, [self._x_parameter,f[i]], P[1], save_timestamp=P[2] and not i)
                     for i in range(len(self._segments)) for P in [["amplitude_%i","arb. unit",True],["phase_%i","rad",False],["real_%i","",False],["imag_%i","",False]]])
        else:
            f = self.Coordinate('frequency', unit='Hz', values=self._freqpoints)
    
            self._prepare_measurement_file(
                    [self.Data("amplitude", [self._x_parameter, f], "arb. unit", save_timestamp=True), self.Data("phase", [self._x_parameter, f], "rad")])
       
        self._pb = Progress_Bar(len(self._x_parameter.values), '2D VNA sweep ' + self.measurement_name, self.vna.get_sweeptime_averages(),
                                dummy=not self.progress_bar)
        
        if self._nop < 10:
            # creates view: plot middle point vs x-parameter, for qubit measurements
            self._views = [self._data_file.add_view("amplitude_midpoint", x=self._x_parameter.hdf_dataset, y=self._datasets['amplitude'],
                                                    view_params=dict(transpose=True, default_trace=self._nop // 2, linecolors=[(200, 200, 100)])),
                self._data_file.add_view("phase_midpoint", x=self._x_parameter.hdf_dataset, y=self._datasets['phase'],
                                         view_params=dict(transpose=True, default_trace=self._nop // 2, linecolors=[(200, 200, 100)]))]
            self._open_qviewkit(['views/amplitude_midpoint', 'views/phase_midpoint'])
        else:
            self._views = []
            self._open_qviewkit(datasets=[] if len(self._segments)>4 else None)
        
        if self._fit_resonator:
            self._resonator = resonator(self._data_file.get_filepath())
        self._measure()
    
    def measure_3D(self):
        """
        measure full window of vna while sweeping x_set_obj and y_set_obj with parameters x_vec/y_vec. sweep over y_set_obj is the inner loop, for every value x_vec[i] all values y_vec are measured.

        optional: measure method to perform the measurement according to landscape, if set
        self.span is the range (in units of the vertical plot axis) data is taken around the specified funtion(s)
        note: make sure to have properly set x,y vectors before generating traces
        """
        if self.landscape.xzlandscape_func:  # The vna limits need to be adjusted, happens in the frequency wrapper
            self._x_parameter.set_function = self.landscape.vna_frequency_wrapper(self._x_parameter.set_function)

        self._dim = 3
        self._measurement_object.measurement_func = 'measure_3D'
        
        self._prepare_measurement_devices()
        if self._segments:
            f =  [self.Coordinate('frequency_%i' % i, unit='Hz', values=self._freqpoints[0 if i == 0 else self._segments[i - 1]:self._segments[i]]) for i in range(len(self._segments))]
            self._prepare_measurement_file(
                    [self.Data(P[0] % i, [self._x_parameter, self._y_parameter,f[i]], P[1], save_timestamp=P[2] and not i)
                     for i in range(len(self._segments)) for P in [["amplitude_%i","arb. unit",True],["phase_%i","rad",False],["real_%i","",False],["imag_%i","",False]]])
        else:
            f = self.Coordinate('frequency', unit='Hz', values=self._freqpoints)
            self._prepare_measurement_file([self.Data("amplitude", [self._x_parameter, self._y_parameter, f], "arb. unit", save_timestamp=True),
                self.Data("phase", [self._x_parameter, self._y_parameter, f], "rad")])
        
        self._open_qviewkit(datasets=[] if len(self._segments)>4 else None)
        
        if self._fit_resonator:
            self._resonator = resonator(self._data_file.get_filepath())
        
        if self.progress_bar:
            if self.landscape.xylandscapes:  # ToDo: This part could be part of the Landscape class
                truth = np.full((len(self._y_parameter.values), len(self._x_parameter.values)), False)  # first, nothing is selected:
                for e in self.landscape.xylandscapes:
                    if not e['blacklist']:
                        truth = np.logical_or(truth, (np.abs(self._y_parameter.values[:, np.newaxis] - e['center_points']) <= e['y_span'] / 2) *  # check y span
                                              (e['x_range'][0] <= self._x_parameter.values) * (self._x_parameter.values <= e['x_range'][1]))  # check x range
                
                for e in self.landscape.xylandscapes:
                    if e['blacklist']:  # exclude blacklisted areas
                        truth = np.logical_and(truth, np.logical_not(
                                (np.abs(self._y_parameter.values[:, np.newaxis] - e['center_points']) <= e['y_span'] / 2) * (
                                    e['x_range'][0] <= self._x_parameter.values) * (self._x_parameter.values <= e['x_range'][1])))
                points = np.sum(truth)
            else:
                points = len(self._x_parameter.values) * len(self._y_parameter.values)
            self._pb = Progress_Bar(points, '3D VNA sweep ' + self.measurement_name, self.vna.get_sweeptime_averages())
        else:
            self._pb = Progress_Bar(1, dummy=True)
        
        self._measure()
    
    def measure_timetrace(self):
        """
        measure method to record a single VNA timetrace, this only makes sense when span is set to 0 Hz!,
        tested only with KEYSIGHT E5071C ENA and its corresponding qkit driver
        LGruenhaupt 11/2016
        """
        #ToDo: move this to regular sweeps
        if self.vna.get_span() > 0:
            raise ValueError(__name__ + ': For timetrace scans, VNA span needs to be 0 Hz')
        if self.vna.get_Average():
            raise ValueError(__name__ + ': For timetrace scans, Averaging needs to be turned off.')
        self.vna.get_centerfreq()
        self._scan_time = True
        
        self._measurement_object.measurement_func = 'measure_timetrace'
        
        self._prepare_measurement_devices()
        
        t = self.Coordinate('time', unit='s', values=np.arange(0, self._nop, 1) * self.vna.get_sweeptime() / (self._nop - 1))
        f = self.Coordinate('frequency', unit='Hz', values=[self.vna.get_centerfreq()])
        iteration = self.Coordinate("trace_number", unit="", values=np.arange(0, self.number_of_timetraces, 1))

        self._prepare_measurement_file([self.Data("amplitude", [iteration, t], "arb. unit", save_timestamp=True), self.Data("phase", [iteration, t], "rad")],
                coords=f)

        self._pb = Progress_Bar(self.number_of_timetraces, 'VNA timetrace ' + self.measurement_name, self.vna.get_sweeptime_averages(),
                                dummy=not self.progress_bar)
        
        qkit.flow.start()
        try:
            """
            loop: x_obj with parameters from x_vec
            """

            for i, x in enumerate(self._x_parameter.values):
                self._x_parameter.set_function(x)
                qkit.flow.sleep(self._x_parameter.wait_time)
                
                self._acquire_log_functions()
                
                data_amp, data_pha = self._acquire_vna_data()
                self._datasets['amplitude'].append(data_amp)
                self._datasets['phase'].append(data_pha)
                qkit.flow.sleep()
                self._pb.iterate()
        finally:
            self._end_measurement()
            self._scan_time = False
    
    def _acquire_vna_data(self):
        if self.averaging_start_ready:
            self.vna.start_measurement()
            if self._scan_time:
                qkit.flow.sleep(self.vna.get_sweeptime(query=False))  # to prevent timeouts in time scan
            elif self.vna.ready():
                logging.debug("VNA STILL ready... Adding delay")
                qkit.flow.sleep(.2)  # just to make sure, the ready command does not *still* show ready
            
            while not self.vna.ready():
                qkit.flow.sleep(min(self.vna.get_sweeptime_averages(query=False) / 11., .2))
        else:
            self.vna.avg_clear()
            qkit.flow.sleep(self._sweeptime_averages)
        
        """ measurement """
        return self.vna.get_tracedata()
    
    def _append(self,amplitude,phase,real=None,imag=None):
        if self._segments:
            for i in range(len(self._segments)):
                self._datasets['amplitude_%i' % i].append(amplitude[0 if i == 0 else self._segments[i - 1]:self._segments[i]])
                self._datasets['phase_%i' % i].append(phase[0 if i == 0 else self._segments[i - 1]:self._segments[i]])
                if real is not None and imag is not None:
                    self._datasets['real_%i' % i].append(real[0 if i == 0 else self._segments[i - 1]:self._segments[i]])
                    self._datasets['imag_%i' % i].append(imag[0 if i == 0 else self._segments[i - 1]:self._segments[i]])
        else:
            self._datasets['amplitude'].append(amplitude)
            self._datasets['phase'].append(phase)
            if real is not None and imag is not None:
                self._datasets['real'].append(real)
                self._datasets['imag'].append(imag)
        
    def _measure(self):
        """
        measures and plots the data depending on the measurement type.
        the measurement loops feature the setting of the objects and saving the data in the .h5 file.
        """
        qkit.flow.start()
        try:
            """
            loop: x_obj with parameters from x_vec
            """
            for ix, x in enumerate(self._x_parameter.values):
                self._x_parameter.set_function(x)
                qkit.flow.sleep(self._x_parameter.wait_time)
                
                self._acquire_log_functions()
    
                if self._dim == 3:
                    for y in self._y_parameter.values:
                        # loop: y_obj with parameters from y_vec (only 3D measurement)
                        if self.landscape.xylandscapes and not self.landscape.perform_measurement_at_point(x, y, ix):
                            # if point is not of interest (not close to one of the functions)
                            data_amp = np.full(int(self._nop), np.NaN, dtype=np.float16)
                            data_pha = np.full(int(self._nop), np.NaN, dtype=np.float16)  # fill with NaNs
                        else:
                            self._y_parameter.set_function(y)
                            qkit.flow.sleep(self._y_parameter.wait_time)
                            if not self.landscape.xzlandscape_func:  # normal scan
                                data_amp, data_pha = self._acquire_vna_data()
                            else:
                                data_amp, data_pha = self.landscape.get_tracedata_xz(x)
                            self._pb.iterate()
                        self._append(data_amp,data_pha)
                        if self._fit_resonator:
                            self._do_fit_resonator()
                        qkit.flow.sleep()
                    """
                    filling of value-box is done here.
                    after every y-loop the data is stored the next 2d structure
                    """
                    [d.next_matrix() for d in self._datasets.values()]
    
                if self._dim == 2:
                    data_amp, data_pha = self._acquire_vna_data()
                    self._append(data_amp, data_pha)
                    
                    if self._fit_resonator:
                        self._do_fit_resonator()
                    self._pb.iterate()
                    qkit.flow.sleep()
        finally:
            self._end_measurement()
    
    def _end_measurement(self):
        super(spectrum, self)._end_measurement()
        if self.averaging_start_ready: self.vna.post_measurement()
    
    def set_resonator_fit(self, fit_resonator=True, fit_function='', f_min=None, f_max=None):
        """
        sets fit parameter for resonator

        fit_resonator (bool): True or False, default: True (optional)
        fit_function (string): function which will be fitted to the data (optional)
        f_min (float): lower frequency boundary for the fitting function, default: None (optional)
        f_max (float): upper frequency boundary for the fitting function, default: None (optional)
        fit types: 'lorentzian','skewed_lorentzian','circle_fit_reflection', 'circle_fit_notch','fano'
        """
        if not fit_resonator:
            self._fit_resonator = False
            return
        self._functions = {'lorentzian': 0, 'skewed_lorentzian': 1, 'circle_fit_reflection': 2, 'circle_fit_notch': 3, 'fano': 5, 'all_fits': 5}
        try:
            self._fit_function = self._functions[fit_function]
        except KeyError:
            logging.error(
                    'Fit function not properly set. Must be either \'lorentzian\', \'skewed_lorentzian\', \'circle_fit_reflection\', \'circle_fit_notch\', \'fano\', or \'all_fits\'.')
        else:
            self._fit_resonator = True
            self._f_min = f_min
            self._f_max = f_max
    
    def _do_fit_resonator(self):
        """
        calls fit function in resonator class
        fit function is specified in self.set_fit, with boundaries f_mim and f_max
        only the last 'slice' of data is fitted, since we fit live while measuring.
        """
        
        if self._fit_function == 0:  # lorentzian
            self._resonator.fit_lorentzian(f_min=self._f_min, f_max=self._f_max)
        if self._fit_function == 1:  # skewed_lorentzian
            self._resonator.fit_skewed_lorentzian(f_min=self._f_min, f_max=self._f_max)
        if self._fit_function == 2:  # circle_reflection
            self._resonator.fit_circle(reflection=True, f_min=self._f_min, f_max=self._f_max)
        if self._fit_function == 3:  # circle_notch
            self._resonator.fit_circle(notch=True, f_min=self._f_min, f_max=self._f_max)
        if self._fit_function == 4:  # fano
            self._resonator.fit_fano(f_min=self._f_min, f_max=self._f_max)
            # if self._fit_function == 5: #all fits
            # self._resonator.fit_all_fits(f_min=self._f_min, f_max = self._f_max)
    
    def set_plot_comment(self, comment):
        """
        Small comment to add at the end of plot pics for more information i.e. good for wiki entries.
        """
        self._plot_comment = comment
    
    def gen_fit_function(self, curve_f, curve_p, x_range=None, span=None, p0=[-1, 0.1, 7], units=''):
        """
        Generates an xy landscape fit for 3D scan. Function is deprecated and calls landscape.generate_fit_function
        curve_f: 'spline', '2pt' 'lin, 'parab', 'hyp', specifies the fit function to be employed
        curve_p: set of points that are the basis for the fit in the format [[x1,x2,x3,...],[y1,y2,y3,...]], frequencies in Hz
        span: specify span for the generated fit_function, if not given default of self.span is used
        x_range: specify x_range for the function in the format [x_min, x_max]
        p0 (optional): start parameters for the fit, must be an 1D array of length 3 ([a,b,c,d]),
           where for the parabula p0[3] will be ignored
        units: set this to 'Hz' in order to avoid large values that cause the fit routine to diverge
        The parabolic function takes the form  y = a*(x-b)**2 + c , where (a,b,c) = p0
        The hyperbolic function takes the form y = sqrt[ a*(x-b)**2 + c ], where (a,b,c) = p0

        adds a trace to landscape
        """
        logging.warning("This function is deprecated. Better to use landscape.generate_fit_function_xy")
        self.landscape.generate_fit_function_xy(curve_f=curve_f, curve_p=curve_p, x_range=x_range, y_span=span, p0=p0, units=units)
    
    def delete_fit_function(self, n=None):
        """
        legacy support for delete_xy_landscape_function. Please use the other function in the future
        delete single fit function n (with 0 being the first one generated) or the complete landscape for n not specified
        :param n: index of single fit function to be deleted
        :return:
        """
        logging.warning('This function is deprecated for clarity purposes.'
                        ' Please use delete_xy_landscape_function in the future')
        self.landscape.delete_landscape_function_xy(n=n)
    
    def plot_fit_function(self):
        """
        legeacy support for landscape.plot_fit_function_xy. Please use the other function in the future.
        Plots the landscape
        :return:
        """
        self.landscape.plot_xy_landscape()


class Landscape:
    """
    Class for landscape scans in combination with spectroscopy. spectrum gnerates an instance of Landscape in __init__
    Use this instance to generate your fit_functions/landscape for your measurement.
    Instance stores the landscape(functions), lets you plot them, checks if points are to be measured and adjusts
    the vna frequencies.
    """
    
    def __init__(self, vna, spec):
        self.vna = vna
        self.spec = spec  # The spectroscopy object
        self.xylandscapes = []  # List containing dicts
        self.xzlandscape_func = None
        self.xz_freqpoints = None
        self.y_span_default = 200e6  # this is for the xy landscape scan, i.e., span of your y_parameter, e.g, mw_frequency
        self.z_span = self.vna.get_span()  # This is for the xz landscape scan i.e. span of vna is adjusted w/ resp to x
    
    def generate_fit_function_xy(self, curve_f, curve_p, x_range=None, y_span=None, blacklist=False, p0=[-1, 0.1, 7], units=''):
        """
        Use this function if you keep your vna span fixed but want to add a landscape to your x and y parameter.
        I.e., for given x parameter not all y parameters are swept but only those that lie within the specified y-span
        around the fit function.
        Adds a trace to landscape
        curve_f: 'spline', 'lin_spline' 'lin, 'parab', 'hyp', 'transmon' specifies the fit function to be employed
        curve_p: set of points that are the basis for the fit in the format [[x1,x2,x3,...],[y1,y2,y3,...]], frequencies in Hz
        x_range: specify x_range for the function in the format [x_min, x_max]
        y_span: specify the span of your y_axis for the generated fit_function,
                if not given default of self.span is used, if negative the span will be excluded from the measurement
                interesting for overlaping regions
        blacklist: Use this parameter to explicitly exclude a section from the measurement (previously done by negative span)
        p0 (optional): start parameters for the fit, must be an 1D array of length 3 ([a,b,c,d]),
           where for the parabula p0[3] will be ignored
        units: set this to 'Hz' in order to avoid large values that cause the fit routine to diverge

        The parabolic function takes the form  y = a*(x-b)**2 + c , where (a,b,c) = p0
        The hyperbolic function takes the form y = sqrt[ a*(x-b)**2 + c ], where (a,b,c) = p0

        """
        if not qkit.module_available("scipy"):
            raise ImportError('scipy not available.')
        
        if units == 'Hz':
            multiplier = 1e9
        else:
            multiplier = 1
        x_fit = np.array(curve_p[0])
        y_fit = np.array(curve_p[1])
        if x_range is None:
            x_range = [self.spec.x_vec[0], self.spec.x_vec[-1]]
        if y_span is None:
            y_span = self.y_span_default
        if y_span < 0:
            y_span = -y_span
            blacklist = True
            logging.warning('Using a negative span in generate_fit_function_xy is deprecated. Please use blacklist = True instead for more clarity.')
        
        try:
            if curve_f == 'lin_spline':
                f = interp1d(x_fit, y_fit)
                center_points = f(self.spec.x_vec)
            elif curve_f == 'spline':
                center_points = UnivariateSpline(x_fit, y_fit)(self.spec.x_vec)
            else:  # curve_fit procedure differs from splines
                if curve_f == 'parab':
                    fit_fct = self.f_parab
                elif curve_f == 'hyp':
                    fit_fct = self.f_hyp
                elif curve_f == 'lin':
                    fit_fct = self.f_lin
                    p0 = p0[:2]
                elif curve_f == 'transmon':
                    fit_fct = self.f_transmon
                else:
                    raise ValueError('function type not known...aborting')
                popt, pcov = curve_fit(fit_fct, x_fit, y_fit / multiplier, p0=p0)
                center_points = multiplier * fit_fct(self.spec.x_vec, *popt)
        except Exception as message:
            print('fit not successful:', message)
        else:
            self.xylandscapes.append({'center_points': center_points, 'y_span': y_span, 'x_range': [np.min(x_range), np.max(x_range)], 'blacklist': blacklist})
    
    def generate_fit_function_xz(self, curve_f, curve_p, z_span=None, p0=[-1, 0.1, 7]):
        """
        Use this function if you want to adjust your vna span with respect to your x-values, useful if you want to
        track your fresonator, e.g., shifting resonance due qubit resonator coupling or fmr measurements. It works
        in a 2d and 3d scan.
        :param curve_f: fit_function (str) such as 'lin_spline', 'spline', 'parab', 'hyp', 'lin'
        :param curve_p: Valeus to fit in the form [[x0,x1,..xn],[z0,z1,..zn]]
        :param z_span: If None, span from VNA is used. If specified, vna span will be updated.
        :param p0: intial fit parameters
        :return:
        """
        if not qkit.module_available("scipy"):
            raise ImportError('scipy not available.')
        
        if curve_p[1][-1] > 1e6:  # test if z_values are given in Hz or GHz
            multiplier = 1e9
        else:
            multiplier = 1
        if z_span is not None:
            self.z_span = z_span
            self.vna.set_span(z_span)
        else:
            self.z_span = self.vna.get_span()
        x_fit = np.array(curve_p[0])
        z_fit = np.array(curve_p[1])
        try:
            if curve_f == 'lin_spline':
                self.xzlandscape_func = interp1d(x_fit, z_fit)
            elif curve_f == 'spline':
                self.xzlandscape_func = UnivariateSpline(x_fit, z_fit)
            else:  # curve_fit procedure differs from splines
                if curve_f == 'parab':
                    fit_fct = self.f_parab
                elif curve_f == 'hyp':
                    fit_fct = self.f_hyp
                elif curve_f == 'lin':
                    fit_fct = self.f_lin
                    p0 = p0[:2]
                else:
                    raise ValueError('function type not known...aborting')
                popt, pcov = curve_fit(fit_fct, x_fit, z_fit / multiplier, p0=p0)
                self.xzlandscape_func = lambda x: multiplier * fit_fct(x, *popt)
            
            self.xz_freqpoints = np.arange(self.xzlandscape_func(self.spec.x_vec[0]) - self.z_span / 2,
                                           self.xzlandscape_func(self.spec.x_vec[-1]) + self.z_span / 2, self.vna.get_span() / self.vna.get_nop())
        except Exception as message:
            print('fit not successful:', message)
    
    def delete_landscape_function_xy(self, n=None):
        """
        delete single xy landscape function n (with 0 being the first one generated)
        or the complete landscape for n not specified
        """
        if n is not None:
            del self.xylandscapes[n]
        else:
            self.xylandscapes = []
    
    def delete_landscape_function_xz(self):
        """
        delets the xz_landscape_function incl. the xz_freqpoints
        :return: None
        """
        self.xzlandscape_func = None
        self.xz_freqpoints = None
        print('xz_landscape deleted')
    
    def plot_xy_landscape(self):
        """
        Plots the xy landscape(s) (for 3D scan, z-axis (vna) is not plotted
        :return:
        """
        if not qkit.module_available("matplotlib"):
            raise ImportError("matplotlib not found.")
        
        if self.xylandscapes:
            for i in self.xylandscapes:
                try:
                    arg = np.where((i['x_range'][0] <= self.spec.x_vec) & (self.spec.x_vec <= i['x_range'][1]))
                    x = self.spec.x_vec[arg]
                    t = i['center_points'][arg]
                    plt.plot(x, t, color='C1')
                    if i['blacklist']:
                        plt.fill_between(x, t + i['y_span'] / 2., t - i['y_span'] / 2., color='C3', alpha=0.5)
                    else:
                        plt.fill_between(x, t + i['y_span'] / 2., t - i['y_span'] / 2., color='C0', alpha=0.5)
                except Exception as e:
                    print(e)
                    print('invalid trace...skip')
            plt.axhspan(np.min(self.spec.y_vec), np.max(self.spec.y_vec), facecolor='0.5', alpha=0.5)
            plt.xlim(np.min(self.spec.x_vec), np.max(self.spec.x_vec))
            plt.show()
        else:
            print('No trace generated. Use landscape.generate_xy_function')
    
    def plot_xz_landscape(self):
        """
        plots the xz landscape, i.e., how your vna frequency span changes with respect to the x vector
        :return: None
        """
        if not qkit.module_available("matplotlib"):
            raise ImportError("matplotlib not found.")
        
        if self.xzlandscape_func:
            y_values = self.xzlandscape_func(self.spec.x_vec)
            plt.plot(self.spec.x_vec, y_values, 'C1')
            plt.fill_between(self.spec.x_vec, y_values + self.z_span / 2., y_values - self.z_span / 2., color='C0', alpha=0.5)
            plt.xlim((self.spec.x_vec[0], self.spec.x_vec[-1]))
            plt.ylim((self.xz_freqpoints[0], self.xz_freqpoints[-1]))
            plt.show()
        else:
            print('No xz funcion generated. Use landscape.generate_xz_function')
    
    def set_y_span(self, y_span, n=None):
        if not self.xylandscapes:
            self.y_span_default = y_span
        else:
            if n is not None:
                self.xylandscapes[n]['y_span'] = y_span
            else:
                for e in self.xylandscapes:
                    e['y_span'] = y_span
    
    def get_y_span(self):
        return self.y_span_default
    
    def get_y_span_list(self):
        return [e['y_span'] for e in self.xylandscapes]
    
    def vna_frequency_wrapper(self, x_set_obj):
        """
        Function wraps the x_set_obj function for a landscape_xy scan such that the vna frequencies are automatically
        adjusted
        :param x_set_obj: as given by the user
        :return: the wrapped function
        """
        
        def vna_wrapper(x):
            start_freq = self.xz_freqpoints[np.argmin(np.abs(self.xz_freqpoints - (self.xzlandscape_func(x) - self.z_span / 2)))]
            self.vna.set_startfreq(start_freq)
            self.vna.set_stopfreq(start_freq + self.z_span)
            x_set_obj(x)
        
        return vna_wrapper
    
    def perform_measurement_at_point(self, x, y, ix):
        """
        Looks if the point is of interest and returns True or False
        :param x: point in the x-vector
        :param y: point in the y-vector
        :param ix: index of x-vector
        :return: measure_bool, whether or not the point should be measured
        """
        # looks strange but works
        measure = False
        for e in self.xylandscapes:
            if np.abs(e['center_points'][ix] - y) <= e['y_span'] / 2 and e['x_range'][0] <= x <= e['x_range'][1]:  # The point is covered by this span
                if e['blacklist']:
                    return False  # if the point is blacklisted anywhere, we don't need to look further
                else:
                    measure = True
        return measure
    
    def get_tracedata_xz(self, x):
        """
        Function to fill the tracedata with NaNs outside the scan region
        :param x: x_value where we perform the scan
        :return:
        """
        amp = np.full_like(self.xz_freqpoints, np.NaN, dtype=np.float16)
        pha = np.full_like(self.xz_freqpoints, np.NaN, dtype=np.float16)
        startarg = np.argmin(np.abs(self.xz_freqpoints - (self.xzlandscape_func(x) - self.z_span / 2)))
        stoparg = startarg + self.vna.get_nop()
        a, p = self.spec._acquire_vna_data()
        amp[startarg:stoparg] = a
        pha[startarg:stoparg] = p
        return amp, pha
    
    def get_freqpoints_xz(self):
        return self.xz_freqpoints
    
    def f_parab(self, x, a, b, c):
        return a * (x - b) ** 2 + c
    
    def f_hyp(self, x, a, b, c):
        """hyperbolic function with the form y = sqrt[ a*(x-b)**2 + c ]"""
        return np.sqrt(a * (x - b) ** 2 + c)
    
    def f_lin(self, x, a, b):
        return a * x + b
    
    def f_transmon(self, x, w_max, L, I_ext, djj):
        """
        Dispersion of a tunable transmon qubit with junction asymmetry.

        Args:
            x:     x-value
            w_max:    Maximum qubit frequency without detuning.
            L:     Oscillation period in current/x-value.
            I_ext: Offset current/x offset.
            djj:   Josephson junction asymmetry (Ic1 - Ic2)/(Ic1 + Ic2)

        Returns:
            Primal transition frequency of a transmon qubit.
        """
        return w_max * (np.abs(np.cos(np.pi / L * (x - I_ext))) * (1 + djj ** 2 * np.tan(np.pi / L * (x - I_ext)) ** 2) ** .5) ** 0.5



# Yokogawa_GS200.py driver for Yokogawa GS200 DC voltag/current source
# adapted from GS820 driver, Lukas Gruenhaupt @KIT 07/2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
import numpy
import struct
import time


class Yokogawa_GS210(Instrument):
    '''
    This is the driver for the Yokogawa GS200 multi channel source measure unit

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Yokogawa_GS210', address='<GBIP address>',
        reset=<bool>)
    '''

    def __init__(self, name, address, reset=False):
        '''
        Initializes the Yokogawa GS200, and communicates with the wrapper.

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
            reset (bool)     : resets to default values

        Output:
            None
        '''
        # Initialize wrapper functions
        logging.info(__name__ + ' : Initializing instrument Yokogawa_GS200')
        Instrument.__init__(self, name, tags=['physical'])

        # Add some global constants
        self._address = address
        #self._numchs = 1
        self.ramp_wait_time = 0.5
        self._visainstrument = visa.instrument(self._address)
        
        #self._visainstrument.write(':SYST:REM')
        

        # Add parameters to wrapper
        
        self.add_parameter('operation_mode',
            flags=Instrument.FLAG_SET,
            type=str, units='')

        self.add_parameter('source_mode',
            flags=Instrument.FLAG_GETSET,
            type=str, units='')

        self.add_parameter('source_range',
            flags=Instrument.FLAG_GETSET ,
            units='', type=str)
            
        self.add_parameter('output',
            flags=Instrument.FLAG_GETSET ,
            units='', type=str)
        
        self.add_parameter('level', 
            flags=Instrument.FLAG_GETSET,
            type=float, units='')

        self.add_parameter('voltage_protection', 
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            type=float, units='V')

        self.add_parameter('current_protection', 
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            type=float, units='A')
            
        self.add_parameter('4W',
            flags=Instrument.FLAG_GETSET ,
            units='', type=str)

        self.add_parameter('ramp_wait_time', 
            flags=Instrument.FLAG_GET,
            type=float, units = 's')
          



        # Add functions to wrapper
        
        self.add_function('reset')
        self.add_function('get_all')
        self.add_function('set_defaults')


        
        if reset:
            self.reset()
        else:
            self.get_all()
            self.set_defaults()
            

# functions
    def reset(self):     
        '''
        Resets instrument to default values

        Input:
            None
    
        Output:
            None
        '''
        logging.debug(__name__ + ' : Resetting instrument')
        self._visainstrument.write('*RST')
        self.get_all()
      
    def get_all(self):
        '''
        Reads all relevant parameters from instrument

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Get all relevant data from device')
        
        
        self.get('source_mode')
        self.get('source_range')
        self.get('output')
        self.get('level')
        self.get('voltage_protection')
        self.get('current_protection')
        self.get('4W')
        self.get('ramp_wait_time')






            
        
        
    def set_defaults(self):
        '''
        Set to driver defaults:
        Operation mode is set to remote
        Source range is +-10 mA
        Source mode is current mode
        output remains unchanged
        voltage protection (max. value) is 30 V
        current protection (max. value) is 10 mA
        The measuring mode is set to 2-wire sense
        '''
        self.set('operation_mode', 'REM')
        self.set('source_mode', 'curr')
        self.set('source_range', '10mA')
        #self.set('output', 'off')
        self.set('voltage_protection', '30')
        self.set('current_protection', '0.01')
        self.set('4W', 'off')
        self.set('slope_time')
        self.ramp_wait_time = 0.5

    

        
 # parameters
 
    def do_set_operation_mode(self, mode):
        '''
        Set operation mode.
        '''    
        if mode in ['REM','LOC']:
            logging.debug('Set mode to %s mode', mode)
            self._visainstrument.write(':SYST:%s' % mode)        
        else:
            logging.error('invalid mode %s' % mode)
            

 
    def do_set_source_mode(self, mode):
        '''
        Set source mode to current regime.
        '''    
        if mode in ['curr','volt']:
            logging.debug('Set mode to %s mode', mode)
            self._set_func_par_value('SOUR', 'FUNC', mode)
        else:
            logging.error('invalid mode %s' % mode)
        self.get_all()

    def do_get_source_mode(self):
        '''
        Get source mode 
        
        Input:
            channel (int) : 
        '''
        logging.debug('Get source mode')
        return self._get_func_par('SOUR', 'FUNC')
        
            
            
    def do_set_source_range(self, val):
        '''
        sets source range: 
            determines the min and max values and the resolution
            - 1 mA  --> +- 1.2 mA  resolution 10 nA (+-30 V load Voltage)
            - 10 mA  --> +- 12 mA  resolution 100 nA (+-30 V load Voltage)
            - 100 mA --> +- 120 mA  resolution 1 umA (+-30 V load Voltage)
            - 200 mA --> +- 200 mA  resolution 1 umA (+-30 V load Voltage)
            
            - 10 mV  --> +- 12 mV  resolution 100 nA (---)
            - 100 mV  --> +- 120 mV  resolution 1 umA (---)
            - 1 V  --> +- 1.2 mV  resolution 1 umA (+-200 mA load Current)
            - 10 V  --> +- 12 mV  resolution 1 umA (+-200 mA load Current)
            - 30 mV  --> +- 32 mV  resolution 1 umA (+-200 mA load Current)

         
        Input:
            val (string)  : Value + Unit 
            example: 10mA

        Output:

        '''
        logging.debug('Set source range to %s' % val)
        self._set_func_par_value('sour', 'rang', val)

            
    def do_get_source_range(self):
        '''
        Get source range 
        '''
        logging.debug('Get source range')
        
        return self._get_func_par('sour', 'rang')
        
    def do_set_output(self, val):
        '''
        Sets outputs of channels in "on" or "off" position.
        
        '''
        if val in ['on','off']:
            logging.debug('Set output to %s' % val)
            self._set_func_par_value('outp', 'stat', val)
        else:
            logging.error('Invalid value %s' % val)
            
    def do_get_output(self):
        '''
        Gets output state
        '''
        logging.debug('Get status of output')
        #ans=self._get_func_par('outp','stat')
        ans = self._visainstrument.ask(':outp?')
        
        if ans=='zero':
            return 'off'
        if int(ans):
            print 'Attention: Output is set on; Be careful!'      #warning if output is on! No fast current changes if coil is attached
            return 'on'
        else:
            return 'off'
        
    def do_set_level(self, val):
        '''
        Set measuring level
        If source mode is 'curr' values are in [mA]
        If source mode is 'volt' values are in [V]
        '''
        logging.debug('Set measuring level')
        #if self.get('output', query = False) == 'off':
        #    self.set('output', 'on')
        mode = self.get('source_mode')
        if mode == 'curr':
            self._set_func_par_value('sour', 'lev', val*10**-3)    #values are given in mA for current mode
        elif mode == 'volt':
            self._set_func_par_value('sour', 'lev', val)
        else:
            print 'error'

        
    def do_get_level(self):
        '''
        Gets source output level
        If source mode is 'curr' values are in [mA]
        If source mode is 'volt' values are in [V]       
        '''
        logging.debug('Get measuring level')
        mode = self.get('source_mode')
        if mode == 'curr':
            return float(self._get_func_par('sour','lev'))*1000     #values are returned in mA for current mode
        elif mode == 'volt':
            return float(self._get_func_par('sour','lev'))
        else:
            return 'error'
        
        
    def do_set_voltage_protection(self, val):
        '''
        Gets voltage protection value
        '''
        logging.debug('Get voltage protection value')
        self._set_func_par_value('SOUR','PROT:VOLT', val)
        
    def do_get_voltage_protection(self):
        '''
        Gets voltage protection value
        '''                
        return float(self._get_func_par('SOUR','PROT:VOLT'))
        
    def do_set_current_protection(self, val):
        '''
        Gets voltage protection value
        '''
        logging.debug('Get voltage protection value')
        self._set_func_par_value('SOUR','PROT:CURR', val)
        
    def do_get_current_protection(self):
        '''
        Gets voltage protection value
        '''                
        return float(self._get_func_par('SOUR','PROT:CURR'))
        
        
    def do_set_4W(self, val):
        '''
        Sets measurement mode to 4-wire or 2-wire mode.
        Value should be on or off. 
        "On" devotes to 4-wire mode. "Off" devotes to 2-wire mode.
        In the instrument appropriate command is ":sens:rem on/off"
        '''
        if val in ['on','off']:
            logging.debug('Set 4W to %s' % val)
            self._set_func_par_value('sens', 'rem', val)
        else:
            logging.error('Invalid value %s' % val)

    def do_get_4W(self):
        '''
        Gets measurement mode
        '''
        logging.debug('Get 4-wire measurement mode')
        ans = self._get_func_par('sens', 'rem')
        
        if int(ans):
            return 'on'
        else:
            return 'off'
            
    def do_get_ramp_wait_time(self):
        '''
        Get ramp time: Parameter is used for a current ramp and determines the waiting time 
        between two subsequent steps.
        '''
        logging.debug('Get ramp time')
        return self.ramp_wait_time

    def do_set_ramp_wait_time(self, val):
        '''
        Get ramp time: Parameter is used for a current ramp and determines the waiting time 
        between two subsequent steps.
        '''
        logging.debug('Set ramp time to %s s' % val)
        self.ramp_wait_time = val
        
        
        
   # core communication
    def _set_func_par_value(self, func, par, val):
        '''
        For internal use only!!
        Changes the value of the parameter for the function specified

        Input:
            ch (int)
            func (string) :
            par (string)  :
            val (depends) :

        Output:
            None
        '''
        string = ':%s:%s %s\r\n' %(func, par, val)
        logging.debug(__name__ + ' : Set instrument to %s' % string)
        self._visainstrument.write(string)

    def _get_func_par(self, func, par):
        '''
        For internal use only!!
        Reads the value of the parameter for the function specified
        from the instrument

        Input:
            ch (int)      :
            func (string) :
            par (string)  :

        Output:
            val (string) :
        '''
        string = ':%s:%s?\r\n' %(func, par)
        ans = self._visainstrument.ask(string)
        logging.debug(__name__ + ' : ask instrument for %s (result %s)' % \
            (string, ans))
        return ans.lower()
        
        
        
        
        
        
    def ramp_current(self,target, step,  wait_time, showvalue=True, outp = False):
        '''
        Ramps the current starting from the actual value to a target value
        Attention: all values are given in mA
        'step' determines the step size
        'wait' determines the sleep time after every step
        'outp' determines if the output is on (True) or off (False) during the ramp
        '''
        
        op = self.get('output')
        start = self.get_level()
        
        
        if outp == True:
            if op == 'off':
                return 'Attention: Output is off! If desired please turn on output manually and repeat operation'  
            elif op == 'on':
                logging.debug('Start current ramp from {} to {} with output on'.format(start,target))
            else: 
                logging.error('Invalid output mode')
        else:
            if op == 'on':
                return  'Attention: Output is on! If desired please turn off output manually and repeat operation'
            elif op == 'off':
                logging.debug('Start current ramp from {} to {} with output off'.format(start,target))
            else: 
                logging.error('Invalid output mode')
                
        if(target < start): step = -step
        elif target == start: return 'Target value {} mA already reached!'.format(target)
        a = numpy.concatenate( (numpy.arange(start, target, step)[1:], [target]) )
        print 'I = {} mA'.format(self.get('level')),
        for i in a:
            if showvalue==True: 
                self.set_level(i)
                print 'I = {} mA'.format(self.get('level')),
            time.sleep(wait_time)   

        
        
        
        
 
 

        

        

        

    def do_get_value(self, channel):
        '''
        Gets measured value
        '''
        logging.debug('Get measured value')
        return float(self._visainstrument.ask(':chan%d:fetc?' % channel))







"""
This instrument contains multiple classes for communicating with a RS232 serial port

The main class handles the qkit integration and stores one of the other classes, 
which each offer a different way of connecting to the serial port. These include:
- Direct connection via USB to SerialPort cable connected to the measurement pc (requires pyserial)
- Via ethernet using a RS232-LAN adapter configured as a TCP server (requires socket)
- Via ethernet using a RasbPi (requires zmq)

The listed site-packages are initially only accessed alongside establishment of their respesctive connection class 
to keep the requirements to the python environment as low as possible

Author: Marius Frohn <uzrfo@student.kit.edu>
Version: 1.0; (07/2024) 
"""

from qkit.core.instrument_base import Instrument
import logging, time

class SerialPortCommunication(Instrument):
    """
    This main class is a wrapper for the underlying actually chosen communication type and handles qkit integration.
    Thus additionally required packages are only imported if the respective connection is indeed demanded. 

    Initialize with
    <SPC_name> = instruments.create('<SPC_name>', 'SerialPortCommunication', *args for creating connection*)
    """
    def __init__(self, name: str, connection_type: str, **connection_config):
        Instrument.__init__(self, name, tags=[])
        self.connection = None
        if connection_type == "Direct":
            self.connection = SerialPortDirect(**connection_config)
        elif connection_type == "Server":
            self.connection = SerialPortServer(**connection_config)
        elif connection_type == "Rasbpi":
            self.connection = SerialPortRasbpi(**connection_config)
        else:
            logging.error("Could not resolve connection type '{}'. Must be 'Direct', 'Server' or 'Rasbpi'".format(connection_type))
            print("Could not resolve connection type '{}'. Must be 'Direct', 'Server' or 'Rasbpi'".format(connection_type))
            raise ValueError
        self.add_function("communicate")
    def communicate(self, message: list[int], expect_return: int) -> list[int]:
        """
        Passes the communication command to the underlying connection type
        """
        if self.connection is None:
            logging.error("Something went wrong during connection creation")
            return []
        else:
            return self.connection.communicate(message, expect_return)


class SerialPortDirect:
    """
    Helper class handling communication via serial port
    Possible port names can be found before initialization with get_ports()

    Baudrate, parity & stopbits are currently hardcoded to fit the needs of IVVI_BiasDAC, 
    feel free to make this class more adjustable, if necessary
    """
    import serial, serial.serialutil

    @staticmethod
    def get_ports():
        """
        Shortcut for serial.tools.list_ports.main()
        Implemented as static function so one can determine to be used port name before instrument creation

        Input:
            None
        
        Output:
            (None): prints information about currently available serial ports
        """
        import serial.tools.list_ports
        serial.tools.list_ports.main()
    
    def __init__(self, port: str = "COM3", timeout: float = 3, timeskip: float = 0.01):
        self.timeskip = timeskip
        try:
            self.ser = self.serial.Serial(port=port, baudrate=115200, parity=self.serial.PARITY_ODD, stopbits=self.serial.STOPBITS_ONE, timeout=timeout, write_timeout=timeout)
        except self.serial.serialutil.SerialException:
            print("Error while trying to open serial port '{}' for IVVI. Port not connected?".format(port))
            raise self.serial.serialutil.SerialException
        self.ser.close()
    
    def __del__(self):
        self.ser.close()
    
    def communicate(self, message: list[int], expect_return: int) -> list[int]:
        """
        Main communication protocol of the serial port connection with integrated error flag interpretation

        Input:
            message (list[int]) : to be sent message
        
        Output:
            (list[int]) : received response, stripped off size & error flag bytes 
        """
        self.ser.open()
        if self.ser.in_waiting > 0:
            # clear input if something still in waiting for whetever reason
            self.ser.read(self.ser.in_waiting)
            time.sleep(self.timeskip)
        self.ser.write(bytes(message))
        time.sleep(self.timeskip)
        readback = []
        for i in range(100):
            readback += list(self.ser.read(self.ser.in_waiting))
            if len(readback) >= expect_return:
                break
            time.sleep(self.timeskip)
        self.ser.close()
        return readback

class SerialPortServer:
    """
    Helper class for accessing a TCP server to do the communication with the serial port.

    For now, this is used for a PUSR rs232-ethernet converter 
    (see https://www.pusr.com/products/1-port-rs232-to-ethernet-converters-usr-tcp232-302.html).
    To configure one of these converters for your desired purpose, download & install the configuration software, 
    then find your device and set it to static IP with what address + port you want and put it into 'TCP server' mode.
    """
    import socket
    def __init__(self, address: str = "129.13.93.50", port: int = 12345, max_message_len: int = 64, timeskip: int = 0.01):
        self.sock = self.socket.socket(self.socket.AF_INET, self.socket.SOCK_STREAM)
        self.sock.connect((address, port))
        self.max_msg_len = max_message_len
        self.timeskip = timeskip
    def __del__(self):
        self.sock.shutdown()
        self.sock.close()
    def communicate(self, message: list[int], expect_return: int) -> list[int]:
        # clear old message 
        self.sock.sendall(bytes(message))
        time.sleep(self.timeskip)
        readback = []
        for i in range(100):
            readback += list(self.sock.recv(self.max_msg_len))
            if len(readback) >= expect_return:
                break
            time.sleep(self.timeskip)
        return readback

class SerialPortRasbpi:
    """
    Helper class handling communication via ethernet using a RasbPi as server

    This mess originated from obsolete version of IVVI_BiasDAC driver, 
    I have no idea if this works or is usable for other cases than IVVI
    """
    import zmq

    def __init__(self, address='10.22.197.115'):
        self._address = address
        self._port = 6543
        self.REQUEST_TIMEOUT = 2500
        self.REQUEST_RETRIES = 3
        self._open_zmq_connection()

    def __del__(self):
        self._close_zmq_connection()

    def _open_zmq_connection(self):
        """
        Connects to the raspberry via zmq/tcp

        Input:
            None

        Output:
            None
        """
        self.context = self.zmq.Context(1)
        print("Connecting to Raspberry Pi")
        self.client = self.context.socket(self.zmq.REQ)
        self.client.connect("tcp://%s:%s" % (self._address,self._port)) # raspi address
        self.poll = self.zmq.Poller()
        self.poll.register(self.client, self.zmq.POLLIN)
        
    def _close_zmq_connection(self):
        """
        Closes the zmq connection

        Input:
            None

        Output:
            None
        """
        logging.debug(__name__ + ' : Closing ethernet connection')
        self.context.term()

    def communicate(self, message: list[int], expect_return: int) -> list[int]:
        """
        Main communication protocol of the ethernet connection with integrated error flag interpretation

        Input:
            message (list[int]) : to be sent message
        
        Output:
            (list[int]) : received response, stripped off size & error flag bytes 
        """
        logging.debug(__name__ + ' : do communication with instrument')
        message = ("{} "*message[0]).format(*message)
        retries_left = self.REQUEST_RETRIES
        while retries_left:
            self.client.send(message)
            expect_reply = True
            while expect_reply:
                socks = dict(self.poll.poll(self.REQUEST_TIMEOUT))
                if socks.get(self.client) == self.zmq.POLLIN:
                    data_out_string = self.client.recv()
                    if not data_out_string:
                        break
                    else:
                        retries_left = 0
                        expect_reply = False
                else:
                    print("No response from server, retrying")
                    # Socket is confused. Close and remove it.
                    self.client.setsockopt(self.zmq.LINGER, 0)
                    self.client.close()
                    self.poll.unregister(self.client)
                    retries_left -= 1
                    if retries_left == 0:
                        print("Server seems to be offline, abandoning")
                        break
                    print("Reconnecting and resending " + message)
                    # Create new connection
                    self._open_zmq_connection()
        return [int(s) for s in data_out_string.split(' ')]
        


# qtLAB preliminary driver for RF frontend PCB
# filename: RFPCBJ.py
# Nej <jdangel@kit.edu>, 2021

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
import zerorpc
import types
import logging
import json
import warnings
from time import sleep
from enum import Enum

class RFPCBJ(Instrument):
    '''
    This is the pre-release python driver for the RF frontend PCB
    Usage:
        Initialise with
        <name> = instruments.create('<name>', "RFPCBJ", address='<TCP/IP>:<port>')
    '''

    def __init__(self, name, address, model = 'RFPCBJ'):
        '''
        Initializes the ZeroRPC Client to communicate with according Server (RPi).
        Input:
            name (string)    : name of the instrument
            address (string) : TCPIP Address including port e.g. "10.22.197.145:4242"
        '''

        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name)
        self._address = "tcp://" + address
        self._model = model
        # self._statusdict = {}
        self._rpc = zerorpc.Client()
        self._rpc.connect(self._address)
        
        self._frequency = 4e9
        self._power = -30
        
        self.set_hf_attenuation(self._power)
        #self.set_frequency

        # Implement parameters
        self.add_parameter('frequency', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=4e9, maxval=12e9,
            units='Hz')
        self.add_parameter('power', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-30, maxval=0,
            units='dBm')
        self.add_parameter('status', type=bool,
            flags=Instrument.FLAG_GETSET)
        
        self.off()
        # Implement functions
        # self.get_all(True)
#######################
#  shiftreg#
######################
    def set_shiftreg_switch(self, boardnr, channel, boolean):
        self._rpc.writeval("SHIFTREG"+str(boardnr), "switches",str(channel) +"," +str(boolean) )
    def set_shiftreg_attenuation(self, boardnr, channel, value):
        self._rpc.writeval("SHIFTREG"+str(boardnr), "attenuation",str(channel) +"," +str(int(2*value)) )
    def shiftreg_flush(self, boardnr, channel, boolean):
        self._rpc.writeval("SHIFTREG"+str(boardnr), "read_flush",str(channel) +"," +str(boolean) )

#####################
#  LOs#
######################

    def set_lo_powerdown(self,boardnr, boolean):
        self._rpc.writeval("LO"+str(boardnr), "power_down", int(boolean))
    def set_lo_init(self,boardnr):
        self._rpc.writeval("LO"+str(boardnr), "init",1)
        self._rpc.readval("LO"+str(boardnr), "read_regs_from_chip")
    def set_lo_frequency(self,boardnr, f):
        self._rpc.writeval("LO"+str(boardnr), "vco_out_frequency", int(f))
    def set_lo_enable_a(self,boardnr, boolean):
        self._rpc.writeval("LO"+str(boardnr), "enable_a", int(boolean))
    def set_lo_enable_b(self,boardnr, boolean):
        self._rpc.writeval("LO"+str(boardnr), "enable_b", int(boolean))
    def set_lo_power_a(self,boardnr, power):
        self._rpc.writeval("LO"+str(boardnr), "output_a_power", int(power))
    def set_lo_power_b(self,boardnr, power):
        self._rpc.writeval("LO"+str(boardnr), "output_b_power", int(power))
    def set_lo_flush(self,boardnr, boolean):
        self._rpc.writeval("LO"+str(boardnr), "write_regs_to_chip", 1)
    def set_lo_read(self,boardnr):
        self._rpc.readval("LO"+str(boardnr), "read_regs_from_chip")


    # initialization related TODO: implement read function (drivers need to be done first)
    # def get_all(self, query=True):
    #     '''
    #     Get Parameters (frequency, amplitude, output on/off, clock reference) of Device as dictionary
    #     Input:
    #         query (bool): Query from device (true) or relay local info (false)
    #     Output:
    #         Dictionary of device
    #     '''
    #     if not query:
    #         return self._statusdict
    #     self._statusdict = json.loads(self._rpc.readall())
    #     self._frequency = float(self._statusdict["frequency"])
    #     self._power = float(self._statusdict["amplitude"])
    #     self._status = bool(int(self._statusdict["rfoutput"]))
    #     refext = bool(int(self._statusdict["reference_int_ext"]))
    #     refint = bool(int(self._statusdict["reference_tcxo_ocxo"]))
    #     if refext:
    #         self._reference = ReferenceSource.EXTERNAL
    #     elif refint:
    #         self._reference = ReferenceSource.OCXO
    #     else:
    #         self._reference = ReferenceSource.TCXO
    #     return self._statusdict

    # def get_diag(self):
    #     '''
    #     Get diagnostics parameters (Firmware Versions, PLL locks status,
    #          Voltage, Temperature) of Device as dictionary
    #     Input:
    #         None
    #     Output:
    #         Dictionary diagnostics parameters
    #     '''
    #     return json.loads(self._rpc.readdiag())

    def set_rflo_powerdown(self, boolean):
        self._rpc.writeval("RFLO", "power_down", int(boolean))

    def set_iflo_powerdown(self, boolean):
        self._rpc.writeval("ZFLO", "power_down", int(boolean))

    def set_txpilot_powerdown(self, boolean):
        self._rpc.writeval("TXPilot", "power_down", int(boolean))

    def set_rxpilot_powerdown(self, boolean):
        self._rpc.writeval("RXPilot", "power_down", int(boolean))


    def set_rflo_enable_a(self, boolean):
        self._rpc.writeval("RFLO", "enable_a", int(boolean))

    def set_iflo_enable_a(self, boolean):
        self._rpc.writeval("ZFLO", "enable_a", int(boolean))

    def set_txpilot_enable_a(self, boolean):
        self._rpc.writeval("TXPilot", "enable_a", int(boolean))

    def set_rxpilot_enable_a(self, boolean):
        self._rpc.writeval("RXPilot", "enable_a", int(boolean))


    def set_rflo_enable_b(self, boolean):
        self._rpc.writeval("RFLO", "enable_b", int(boolean))

    def set_iflo_enable_b(self, boolean):
        self._rpc.writeval("ZFLO", "enable_b", int(boolean))

    def set_txpilot_enable_b(self, boolean):
        self._rpc.writeval("TXPilot", "enable_b", int(boolean))

    def set_rxpilot_enable_b(self, boolean):
        self._rpc.writeval("RXPilot", "enable_b", int(boolean))


    def set_rflo_frequency(self, value):
        self._rpc.writeval("RFLO", "vco_out_frequency", int(value))
        self._rpc.writeval("RFLO", "recalibrate", 1)

    def set_iflo_frequency(self, value):
        self._rpc.writeval("ZFLO", "vco_out_frequency", int(value))
        self._rpc.writeval("ZFLO", "recalibrate", 1)
        self._rpc.writeval("Demodulator", "lo_match_mhz", int(value/1e6))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        

    def set_txpilot_frequency(self, value):
        self._rpc.writeval("TXPilot", "vco_out_frequency", int(value))
        self._rpc.writeval("TXPilot", "recalibrate", 1)

    def set_rxpilot_frequency(self, value):
        self._rpc.writeval("RXPilot", "vco_out_frequency", int(value))
        self._rpc.writeval("RXPilot", "recalibrate", 1)


    def set_rflo_output_a_power(self, value):
        self._rpc.writeval("RFLO", "output_a_power", int(value))

    def set_iflo_output_a_power(self, value):
        self._rpc.writeval("ZFLO", "output_a_power", int(value))

    def set_txpilot_output_a_power(self, value):
        self._rpc.writeval("TXPilot", "output_a_power", int(value))

    def set_rxpilot_output_a_power(self, value):
        self._rpc.writeval("RXPilot", "output_a_power", int(value))


    def set_rflo_output_b_power(self, value):
        self._rpc.writeval("RFLO", "output_b_power", int(value))

    def set_iflo_output_b_power(self, value):
        self._rpc.writeval("ZFLO", "output_b_power", int(value))

    def set_txpilot_output_b_power(self, value):
        self._rpc.writeval("TXPilot", "output_b_power", int(value))

    def set_rxpilot_output_b_power(self, value):
        self._rpc.writeval("RXPilot", "output_b_power", int(value))


    def set_linopt(self, value):
        if (value > 3000):
            print("Linopt would exceed safe limits!\n Defaulting to 3V55")
            value = 3000
        self._rpc.setlinopt(value)

    def set_hf_attenuation(self, value1, value2 = 0):
        self._rpc.SetHFAtt(value1, value2)
        

    def set_demod_attenuation(self, value):
        self._rpc.writeval("Demodulator", "attenuation", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
#    def set_demod_band(self,value):
#        self._rpc.writeval("Demodulator", "band", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
#    def set_demod_dc_offset_i(self, value):
#        self._rpc.writeval("Demodulator", "dc_offset_i", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
#    
#    def set_demod_dc_offset_q(self, value):
#        self._rpc.writeval("Demodulator", "dc_offset_q", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)

    def set_demod_gain(self, value):
        self._rpc.writeval("Demodulator", "gain", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
    def set_demod_gain_error(self, value):
        self._rpc.writeval("Demodulator", "gain_error", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
    #def set_demod_hd2i(self, value):
    #def set_demod_hd2q(self, value):
    #def set_demod_hd3i(self, value):
    #def set_demod_hd3q(self, value):
    #def set_demod_im2(self, value):
    #def set_demod_im3(self, value):
    #def set_demod_im3i(self, value):
    #def set_demod_im3q(self, value):
    #def set_demod_ip3(self, value):
    #def set_demod_lo_bias(self, value): 
#    def set_demod_lo_match_c1(self, value):
#        self._rpc.writeval("Demodulator", "lo_match_c1", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
#    
#    def set_demod_lo_match_c2(self, value):
#        self._rpc.writeval("Demodulator", "lo_match_c2", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
#        
#    def set_demod_lo_match_inductor(self, value):
#        self._rpc.writeval("Demodulator", "lo_match_inductor", int(value))
#        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
    def set_demod_phase_error(self, value):
        self._rpc.writeval("Demodulator", "phase_error", int(value))
        self._rpc.writeval("Demodulator", "write_regs_to_chip", 1)
        
    #def set_demod_rf_switch(self, value):

    def do_get_frequency(self, query=True):
        '''
        Get frequency of device
        Input:
            query (bool): Refresh parameters from device memory if True
        Output:
            microwave frequency (Hz)
        '''
        # TODO: read from LOs to calculate band
        return self._frequency
        
    def do_set_frequency(self, frequency, offset = 100e6, method="rfsweep"):
        '''
        Set frequency of device.
        Default BB frequency is 100MHz.
        Default method is sweeping the RF LO.
        Input:
            freq (float) : Frequency in Hz
        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting frequency to %s Hz' % (frequency*1.0e9))
        
        iflo = 4e9 #Filtered 1,8 GHz to 5 GHz
        rflo = 8e9 #Filtered 3,8 GHz to 8,4 GHz
        
        if(method == "ifsweep"):
            # RF LO stationary, IF variable, USB
            # This can reduce spurs from RF LO harmonics.
            # IF drops by 10dB towards 4 GHz though.
            if(frequency <= 6e9 or frequency >= 10e9):
                rflo = 8e9
            elif(frequency > 6e9 and frequency <= 8e9):
                rflo = 4e9
            elif(frequency > 8e9 and frequency < 10e9):
                rflo = 6e9
            else:
                print("No valid LO configuration found!")
            iflo = abs(frequency - rflo + offset)
        elif(method == "ifsweep_extend"):
            # IF level drops even further for 5 GHz.
            if(frequency <= 6.5e9):
                rflo = 8.5e9
            elif(frequency > 6.5e9 and frequency < 9e9):
                rflo = 4e9
            elif(frequency >= 9e9 and frequency < 12e9):
                rflo = 7e9
            else:
                print("No valid LO configuration found!")
            iflo = abs(frequency - rflo + offset) # maybe + 100MHz?
        elif(method == "rfsweep"):
            # IF LO stationary 2.3G => highest amplitude, RF variable
            # Low IF and/or RF frequencyies cause spurs!
            iflo = 2.3e9 - offset
            if(frequency <= 6.1e9):
                rflo = frequency + iflo 
            elif(frequency > 6.1e9 and frequency < 10.7e9):
                rflo = frequency - iflo
            elif(frequency >= 10.7):
                rflo = 8.4e9
                iflo = frequency - rflo 
            else:
                print("No valid LO configuration:")
            iflo += offset
        else:
            print("Please select linear combination method!")
        
        self.set_iflo_frequency(iflo)
        self.set_rflo_frequency(rflo)
        self._frequency = frequency

    def do_get_power(self, query=True):
        '''
        Get output power of device
        Input:
            query (bool): Refresh parameters from device memory if True
        Output:
            microwave power (dBm)
        '''
        #TODO: implement read
        return self._power

    def do_set_power(self, power = 0):
        '''
        Sets 0 to 30 dB RF attenuation.
        Output level depends on modulator input or pilot power
        (usually around -10 to 0 dBm max.)
        Input:
            power (float) : "Power" in dBm
        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting power to %s dBm' % power)
        self._power = power
        self.set_hf_attenuation(abs(power))

    def do_get_status(self, query=True):
        '''
        Get status of output channel
        Input:
            query (bool): Refresh parameters from device memory if True
        Output:
            True (on) or False (off)
        '''
        #self.get_all(query)
        # TODO: implement read
        return self._status

    def do_set_status(self, status):
        '''
        Set status of output (Bool)
        Input:
            status (bool): enable LOs or disable everything
        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting status to "%s"' % status)
        if status:
            self.set_rflo_enable_a(1)
            self.set_rflo_enable_b(1)
            self.set_rflo_output_a_power(5)

            self.set_iflo_enable_a(1)
            self.set_iflo_enable_b(1)
            self.set_iflo_output_a_power(5)
            
            self.set_rflo_powerdown(0)
            self.set_iflo_powerdown(0)
        else:
            self.set_rflo_powerdown(1)
            self.set_iflo_powerdown(1)
            self.set_txpilot_powerdown(1)
            self.set_rxpilot_powerdown(1)
        self._status = bool(status)


    #shortcuts
    def off(self):
        '''Turn everything Off'''
        self.set_status(0)

    def on(self):
        '''Turn LOs On'''
        self.set_status(1)



# filename: Virtual_Coil.py
# Jochen Braumueller <jochen.braumueller@kit.edu>, 11/2013
# updates: 02/2015, 06/2016, 03/2019
# virtual coil instrument to be used as a wrapper for the DELFT tunnel electronics driver IVVI
# use: 
#  - set current range of downstream current source
#  - current unit is mA, independent of the specified and set range

# import and usage
"""
IVVI = qkit.instruments.create('IVVI', 'IVVIDIG_main', address='COM6') #COM6 corresponds to adapter D in the UFO setup
vcoil = qkit.instruments.create('vcoil', 'Virtual_Coil', dacs = [5])   #dacs to be used in list form [dac(ch1), dac(ch2), dac(ch3), ...]
vcoil.init()   #optional
vcoil.set_c_range('1m')
vcoil.set_current(1.0)
"""

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit
from qkit.core.instrument_base import Instrument
import logging
import numpy as np
import time
from qkit.gui.notebook.Progress_Bar import Progress_Bar



DAC_ROUT = [5]   #dac port(s) to be used, passed as a list during instanciating
dac_val = {'100m':2,'20m':1.30103,'10m':1,'1m':0,'100u':-1,'10u':-2,'1u':-3,'100n':-4,'10n':-5,'1n':-6}   #DELFT electronics dictionary

class Virtual_Coil(Instrument):
    def __init__(self, name, dacs=DAC_ROUT, dacranges=None, ivvi_instrument=None):
        """
        init virtual instrument object
        Inputs:
            - name: qt instrument name
            - dacs: list of routes of channels in use, e.g. [5] or [5,9,2]
            - dacranges: list of length len(dacs) indicating the dac ranges set at the device, list entries one of 'bi' for -2000,2000, 'pos' for 0,4000, 'neg' for -4000,0
        """
        Instrument.__init__(self, name, tags=['virtual'])

        if ivvi_instrument is None:
            ins_list = qkit.instruments.get_instruments_by_type('IVVI') + qkit.instruments.get_instruments_by_type(
                    'IVVIDIG_main') + qkit.instruments.get_instruments_by_type('IVVIDIG_main_eth')
            if len(ins_list) > 1:
                raise ImportError("I found %i instruments matching IVVI." % (len(ins_list)))
            else:
                self.IVVI = ins_list[0]
        else:
            self.IVVI = ivvi_instrument
            
        self.dacs = dacs
        try:
            self.channels = len(self.dacs)
        except TypeError:
            logging.error('Error reading dac ports to be used. Import as list.')
        if dacranges == None:
            dacranges = ['bi'] * self.channels
        else:
            if len(dacranges) != self.channels:
                logging.error('dacranges not specified properly')
        limits_dict = {'bi':[-2000,2000],'pos':[0,4000],'neg':[-4000,0]}
        self.range_limits = [limits_dict[dacranges[i]] for i in range(self.channels)]
        # print(self.range_limits)
        if self.channels == 1:
            self.add_parameter('current', type=float, flags=Instrument.FLAG_GETSET, units='mA')
            self.add_parameter('range', type=str, flags=Instrument.FLAG_GETSET)
        elif self.channels > 1 and self.channels <= 16:
            self.add_parameter('current', type=float, flags=Instrument.FLAG_GETSET, units='mA',
                               channels=(1, self.channels), channel_prefix='ch%d_')
            self.add_parameter('range', type=str, flags=Instrument.FLAG_GETSET,
                               channels=(1, self.channels), channel_prefix='ch%d_')
        else:
            logging.error('incorrect number of channels specified')

        self.c_ranges = ['1m'] * self.channels

    def do_set_range(self, range, channel=1):
        """
        set the current range of channel ch
        """
        if range in dac_val:
            self.c_ranges[channel - 1] = range
        else:
            logging.error('Invalid current range.')

    def do_get_range(self, channel=1):
        """
        get the current range of channel ch
        """
        return self.c_ranges[channel - 1]

    def do_set_current(self, current, channel=1, verbose=True):  # current in mA
        """
        set current of channel ch
        """
        try:
            val = np.round(-0.5 * (self.range_limits[channel - 1][0] + self.range_limits[channel - 1][-1]) + (
                        current * 1000 * np.power(10., -dac_val[self.c_ranges[channel - 1]])), 10)  # dac value in mV
            if val < -2000 or val > 2000:
                raise ValueError('DAC range limits exceeded')
            else:
                if 10. * val != np.round(10. * val) and verbose:
                    logging.warning('16 bit resolution may be visible for the current value you are attempting to set.')

                val = np.power(10., -dac_val[self.c_ranges[channel - 1]]) * current * 1000  # dac is in mV
                self.IVVI.set_dac(self.dacs[channel - 1], val)
        except IndexError as detail:
            logging.error('Electronics might be disconnected.')
            raise detail
            
    def ramp_to(self, target = 0, ch = 1, steps = 100, dt = 0.05,show_progressbar=True):
        """
        ramp current to target
        Inputs:
            - target: target current value
            - ch: channel index (1..)
            - steps: number of steps
            - dt: wait time between two steps
        """
        if show_progressbar: p = Progress_Bar(steps,'Ramping current')   #init progress bar
        for c in np.linspace(self.do_get_current(ch),target,steps):
            self.do_set_current(c,ch,verbose=False)
            if show_progressbar: p.iterate("%.3gmA"%c)
            time.sleep(dt)
        self.do_set_current(c, ch, verbose=True)

    def do_get_current(self, channel=1):
        """
        get current of channel ch
        """
        val = float(self.IVVI.get_dac(self.dacs[channel - 1]) + 0.5 * (
                    self.range_limits[channel - 1][0] + self.range_limits[channel - 1][-1])) / 1000  # dac value in mV
        # print(val)
        try:
            return val * np.power(10., dac_val[self.c_ranges[channel - 1]])
        except IndexError as detail:
            logging.error('Electronics might be disconnected.')
            raise detail
        except Exception as detail:
            logging.error(detail)
